import copy

import numpy as np
import torch
import torch.nn as nn

from dingo.core.utils import torchutils
from dingo.core.nn.enets import create_enet_with_projection_layer_and_dense_resnet

from dingo.core.nn.enets import DenseResidualNet


class ContinuousFlow(nn.Module):
    """
    A continuous normalizing flow network. It defines a time-dependent vector field on
    the parameter space (score or flow), which optionally depends on additional context
    information.

    v = v(f(t, theta), g(context))

    This class combines the network v for the continuous flow itself, as well as embedding
    networks f, g, for the context and parameters, respectively.

    The parameters and context can optionally be provided as gated linear unit (GLU)
    context to the main network, rather than as the main input to the network. For a
    DenseResidualNet, this context is input repeatedly via GLUs, for each residual block.
    """

    def __init__(
        self,
        continuous_flow_net: nn.Module,
        context_embedding_net: nn.Module = torch.nn.Identity(),
        theta_embedding_net: nn.Module = torch.nn.Identity(),
        context_with_glu: bool = False,
        theta_with_glu: bool = False,
    ):
        """
        Parameters
        ----------
        continuous_flow_net: nn.Module
            Main network for the continuous flow.
        context_embedding_net: nn.Module = torch.nn.Identity()
            Embedding network for the context information (e.g., observed data).
        theta_embedding_net: nn.Module = torch.nn.Identity()
            Embedding network for the parameters.
        context_with_glu: bool = False
            Whether to provide context as GLU or main input to the continuous_flow_net.
        theta_with_glu: bool = False
            Whether to provide theta (and t) as GLU or main input to the
            continuous_flow_net.
        """
        super(ContinuousFlow, self).__init__()
        self.continuous_flow_net = continuous_flow_net
        self.context_embedding_net = context_embedding_net
        self.theta_embedding_net = theta_embedding_net
        self.theta_with_glu = theta_with_glu
        self.context_with_glu = context_with_glu

        self._use_cache = None
        self._cached_context = None
        self._cached_context_embedding = None

    @property
    def use_cache(self):
        # unless set explicitly, use_cache is True in eval mode and False in train mode
        if self._use_cache is not None:
            return self._use_cache
        else:
            return not self.training

    @use_cache.setter
    def use_cache(self, value):
        self._use_cache = value

    def _update_cached_context(self, *context: torch.Tensor):
        """
        Update the cache for *context. This sets new values for self._cached_context and
        self._cached_context_embedding if self._cached_context != context.
        """
        try:
            # This may fail when batch size of context and _cached_context is different
            # (but both > 1).
            if (
                self._cached_context is not None
                and len(self._cached_context) == len(context)
                and all([(x == y).all() for x, y in zip(self._cached_context, context)])
            ):
                return
        except RuntimeError:
            pass
        # if all tensors in batch are the same: do forward pass with batch_size 1
        if all([(x == x[:1]).all() for x in context]):
            self._cached_context = tuple(x[:1] for x in context)
            self._cached_context_embedding = self.context_embedding_net(
                *self._cached_context
            ).detach()

        else:
            self._cached_context = context
            self._cached_context_embedding = self.context_embedding_net(
                *self._cached_context
            ).detach()

    def _get_cached_context_embedding(self, batch_size):
        if self._cached_context_embedding.size(0) == 1:
            return self._cached_context_embedding.repeat(
                batch_size,
                *[1 for _ in range(len(self._cached_context_embedding.shape) - 1)],
            )
        return self._cached_context_embedding

    def forward(self, t, theta, *context):
        # embed theta (self.embedding_net_theta might just be identity)
        t_and_theta_embedding = torch.cat((t.unsqueeze(1), theta), dim=1)
        t_and_theta_embedding = self.theta_embedding_net(t_and_theta_embedding)
        # for unconditional forward pass
        if len(context) == 0:
            assert not self.theta_with_glu
            return self.continuous_flow_net(t_and_theta_embedding)

        # embed context (self.context_embedding_net might just be identity)
        if not self.use_cache:
            context_embedding = self.context_embedding_net(*context)

        else:
            self._update_cached_context(*context)
            context_embedding = self._get_cached_context_embedding(
                batch_size=len(context[0])
            )

        if len(t_and_theta_embedding.shape) != 2 or len(context_embedding.shape) != 2:
            raise NotImplementedError()

        # a = context_embedding and b = t_and_theta_embedding now need to be provided
        # to the continuous flow network, which predicts a vector field as a function
        # of a and b. The flow network has two entry points: the normal input to the
        # feedforward network (first argument in forward pass) and via a glu between
        # the residual blocks (second argument in forward pass, optional). The flags
        # self.theta_with_glu and self.context_with_glu specify whether we use the
        # first entrypoint (= False) or the second (= True).
        if self.context_with_glu and self.theta_with_glu:
            main_input = torch.Tensor([])
            glu_context = torch.cat((context_embedding, t_and_theta_embedding), dim=1)
        elif not self.context_with_glu and not self.theta_with_glu:
            main_input = torch.cat((context_embedding, t_and_theta_embedding), dim=1)
            glu_context = None
        elif self.context_with_glu:
            main_input = t_and_theta_embedding
            glu_context = context_embedding
        else:  # if self.theta_with_glu:
            main_input = context_embedding
            glu_context = t_and_theta_embedding

        if glu_context is None:
            return self.continuous_flow_net(main_input)
        else:
            return self.continuous_flow_net(main_input, glu_context)


def create_cf(
    posterior_kwargs: dict, embedding_kwargs: dict = None, initial_weights: dict = None
):
    """
    Build a continuous flow based on settings dictionaries.

    Parameters
    ----------
    posterior_kwargs: dict
        Settings for the flow. This includes the settings for the parameter embedding.
    embedding_kwargs: dict
        Settings for the context embedding network.
    initial_weights: dict
        Initial weights for the embedding network (of SVD projection type).

    Returns
    -------
    nn.Module
        Neural network for the continuous flow.
    """
    theta_dim = posterior_kwargs["input_dim"]
    context_dim = posterior_kwargs["context_dim"]

    # get embeddings modules for context
    if embedding_kwargs is not None:
        context_embedding_kwargs = copy.deepcopy(embedding_kwargs)
        if initial_weights is not None:
            context_embedding_kwargs["V_rb_list"] = initial_weights["V_rb_list"]
        elif "V_rb_list" not in context_embedding_kwargs:
            context_embedding_kwargs["V_rb_list"] = None

        context_embedding = create_enet_with_projection_layer_and_dense_resnet(
            **context_embedding_kwargs
        )
    else:
        context_embedding = torch.nn.Identity()

    # get embeddings modules for theta (which is actually cat(t, theta))
    if "theta_embedding_kwargs" in posterior_kwargs:
        theta_embedding = get_theta_embedding_net(
            posterior_kwargs["theta_embedding_kwargs"],
            input_dim=theta_dim + 1,
        )
    else:
        theta_embedding = torch.nn.Identity()

    # get output dimensions of embedded context and theta
    theta_with_glu = posterior_kwargs.get("theta_with_glu", False)
    context_with_glu = posterior_kwargs.get("context_with_glu", False)
    embedded_theta_dim = theta_embedding(torch.zeros(10, theta_dim + 1)).shape[1]

    glu_dim = theta_with_glu * embedded_theta_dim + context_with_glu * context_dim
    input_dim = embedded_theta_dim + context_dim - glu_dim
    if glu_dim == 0:
        glu_dim = None

    activation_fn = torchutils.get_activation_function_from_string(
        posterior_kwargs["activation"]
    )
    continuous_flow_net = DenseResidualNet(
        input_dim=input_dim,
        output_dim=theta_dim,
        hidden_dims=posterior_kwargs["hidden_dims"],
        activation=activation_fn,
        dropout=posterior_kwargs["dropout"],
        batch_norm=posterior_kwargs["batch_norm"],
        context_features=glu_dim,
    )

    model = ContinuousFlow(
        continuous_flow_net,
        context_embedding,
        theta_embedding,
        theta_with_glu=posterior_kwargs.get("theta_with_glu", False),
        context_with_glu=posterior_kwargs.get("context_with_glu", False),
    )
    return model


def get_theta_embedding_net(embedding_kwargs: dict, input_dim):
    if "encoding" in embedding_kwargs:
        input_dim = get_dim_positional_embedding(
            embedding_kwargs["encoding"], input_dim
        )
        positional_encoding = PositionalEncoding(
            nr_frequencies=embedding_kwargs.get("frequencies", 0),
            encode_all=embedding_kwargs.get("encode_all"),
        )
    else:
        positional_encoding = torch.nn.Identity()

    if "embedding_net" in embedding_kwargs:
        activation_fn = torchutils.get_activation_function_from_string(
            embedding_kwargs["embedding_net"]["activation"]
        )
        embedding_net = DenseResidualNet(
            input_dim=input_dim,
            output_dim=embedding_kwargs["embedding_net"]["output_dim"],
            hidden_dims=embedding_kwargs["embedding_net"]["hidden_dims"],
            activation=activation_fn,
            dropout=embedding_kwargs["embedding_net"].get("dropout", 0.0),
            batch_norm=embedding_kwargs["embedding_net"].get("batch_norm", True),
        )
    else:
        embedding_net = torch.nn.Identity()

    return torch.nn.Sequential(positional_encoding, embedding_net)


def get_dim_positional_embedding(encoding: dict, input_dim: int):
    if encoding.get("encode_all"):
        return (1 + 2 * encoding["frequencies"]) * input_dim
    return 2 * encoding["frequencies"] + input_dim

class PositionalEncoding(nn.Module):
    """
    Implements positional encoding as commonly used in transformer architectures.
    
    Positional encoding introduces a way to inject information about the order of 
    the input data (e.g., sequence positions) into a neural network that otherwise 
    lacks a sense of position due to its permutation-invariant nature. This class 
    computes sinusoidal encodings based on the position of each element in the input 
    and concatenates them with the original input features.

    Attributes
    ----------
    frequencies : torch.Tensor
        A tensor containing the frequencies used to calculate the sinusoidal components.
        The frequencies are powers of 2, scaled by the base frequency.
    encode_all : bool
        Determines whether the positional encoding is applied to all features of the input
        or only the first feature (e.g., time component).
    base_freq : float
        The base frequency used to scale the sinusoidal components, defaulting to `2 * pi`.

    Parameters
    ----------
    nr_frequencies : int
        The number of sinusoidal frequencies to compute. This determines the dimensionality
        of the positional encoding for each input feature.
    encode_all : bool, optional (default=True)
        If True, the positional encoding is computed for all features in the input. 
        Otherwise, it is computed only for the first feature (e.g., the time dimension).
    base_freq : float, optional (default=2 * np.pi)
        The base frequency used for sinusoidal encoding.

    Methods
    -------
    forward(t_theta)
        Computes the positional encoding for the input tensor `t_theta` and concatenates 
        it with the original input features.
        - If `encode_all` is True, the positional encoding is computed for all features.
        - If `encode_all` is False, the positional encoding is applied only to the first
          feature, such as time, while other features remain unchanged.
    """
    def __init__(self, nr_frequencies, encode_all=True, base_freq=2 * np.pi):
        super(PositionalEncoding, self).__init__()
        frequencies = base_freq * torch.pow(
            2 * torch.ones(nr_frequencies), torch.arange(0, nr_frequencies)
        ).view(1, 1, nr_frequencies)
        self.register_buffer("frequencies", frequencies)
        self.encode_all = encode_all

    def forward(self, t_theta):
        """
        Computes and concatenates positional encodings with the input tensor.

        Parameters
        ----------
        t_theta : torch.Tensor
            Input tensor of shape (batch_size, input_dim), where `input_dim` is the
            dimensionality of the input features.

        Returns
        -------
        torch.Tensor
            A tensor containing the input features concatenated with the positional
            encodings. The output shape will be:
            - (batch_size, input_dim + 2 * nr_frequencies) if `encode_all` is True.
            - (batch_size, input_dim + 2 * nr_frequencies) if `encode_all` is False,
              but positional encodings are computed only for the first input feature.
        """
        batch_size = t_theta.size(0)
        if self.encode_all:
            x = t_theta.view(batch_size, -1, 1) * self.frequencies
        else:
            x = t_theta[:, 0:1].view(batch_size, 1, 1) * self.frequencies
        cos_enc, sin_enc = torch.cos(x).view(batch_size, -1), torch.sin(x).view(
            batch_size, -1
        )
        return torch.cat((t_theta, cos_enc, sin_enc), dim=1)



"""These functions are used for transforming between J and L0 frames."""
import numpy as np
import lal

def rotate_z(angle, vx, vy, vz):
    vx_new = vx * np.cos(angle) - vy * np.sin(angle)
    vy_new = vx * np.sin(angle) + vy * np.cos(angle)
    return vx_new, vy_new, vz


def rotate_y(angle, vx, vy, vz):
    vx_new = vx * np.cos(angle) + vz * np.sin(angle)
    vz_new = -vx * np.sin(angle) + vz * np.cos(angle)
    return vx_new, vy, vz_new


def get_JL0_euler_angles(p, wfg, spin_conversion_phase=None):
    p = p.copy()
    p["f_ref"] = wfg.f_ref
    phase = p["phase"]

    if spin_conversion_phase is not None:
        p["phase"] = 0.0
    p_lal = wfg._convert_parameters(p)
    p_lal = list(p_lal)

    m1, m2 = p_lal[0:2]
    chi1x, chi1y, chi1z, chi2x, chi2y, chi2z = p_lal[2:8]
    iota = p_lal[9]

    m1 = m1 / lal.MSUN_SI
    m2 = m2 / lal.MSUN_SI

    m = m1 + m2
    eta = m1 * m2 / m ** 2
    v0 = (m * lal.MTSUN_SI * np.pi * wfg.f_ref) ** (1 / 3)

    # All quantities in L0 frame, unless indicated.
    s1x = m1 * m1 * chi1x
    s1y = m1 * m1 * chi1y
    s1z = m1 * m1 * chi1z
    s2x = m2 * m2 * chi2x
    s2y = m2 * m2 * chi2y
    s2z = m2 * m2 * chi2z

    delta = np.sqrt(1 - 4 * eta)
    m1_prime = (1 + delta) / 2
    m2_prime = (1 - delta) / 2
    Sl = m1_prime ** 2 * chi1z + m2_prime ** 2 * chi2z
    Sigmal = chi2z * m2_prime - chi1z * m1_prime

    # This calculation of the orbital angular momentum is taken from Appendix G.2 of PRD 103, 104056 (2021).
    # It may not align exactly with the various XPHM PrecVersions, but the error should not be too big.
    Lmag = (m * m * eta / v0) * (
        1
        + v0 * v0 * (1.5 + eta / 6)
        + (27 / 8 - 19 * eta / 8 + eta ** 2 / 24) * v0 ** 4
        + (
            7 * eta ** 3 / 1296
            + 31 * eta ** 2 / 24
            + (41 * np.pi ** 2 / 24 - 6889 / 144) * eta
            + 135 / 16
        )
        * v0 ** 6
        + (
            -55 * eta ** 4 / 31104
            - 215 * eta ** 3 / 1728
            + (356035 / 3456 - 2255 * np.pi ** 2 / 576) * eta ** 2
            + eta
            * (
                -64 * np.log(16 * v0 ** 2) / 3
                - 6455 * np.pi ** 2 / 1536
                - 128 * lal.GAMMA / 3
                + 98869 / 5760
            )
            + 2835 / 128
        )
        * v0 ** 8
        + (-35 * Sl / 6 - 5 * delta * Sigmal / 2) * v0 ** 3
        + ((-77 / 8 + 427 * eta / 72) * Sl + delta * (-21 / 8 + 35 * eta / 12) * Sigmal)
        * v0 ** 5
    )

    Jx = s1x + s2x
    Jy = s1y + s2y
    Jz = Lmag + s1z + s2z

    Jnorm = np.sqrt(Jx * Jx + Jy * Jy + Jz * Jz)
    Jhatx = Jx / Jnorm
    Jhaty = Jy / Jnorm
    Jhatz = Jz / Jnorm

    # The calculation of the Euler angles is described in Appendix C of PRD 103, 104056 (2021).
    theta_JL0 = np.arccos(Jhatz)
    phi_JL0 = np.arctan2(Jhaty, Jhatx)

    Nx = np.sin(iota) * np.cos(np.pi / 2 - phase)
    Ny = np.sin(iota) * np.sin(np.pi / 2 - phase)
    Nz = np.cos(iota)

    # Rotate N into J' frame.
    Nx_Jp, Ny_Jp, Nz_Jp = rotate_y(-theta_JL0, *rotate_z(-phi_JL0, Nx, Ny, Nz))

    kappa = np.arctan2(Ny_Jp, Nx_Jp)

    alpha_0 = np.pi - kappa
    beta_0 = theta_JL0
    gamma_0 = np.pi - phi_JL0

    return alpha_0, beta_0, gamma_0


def convert_J_to_L0_frame(hlm_J, p, wfg, spin_conversion_phase=None):
    alpha_0, beta_0, gamma_0 = get_JL0_euler_angles(
        p, wfg, spin_conversion_phase=spin_conversion_phase
    )

    hlm_L0 = {}
    for (l, m), hlm in hlm_J.items():
        for mp in range(-l, l + 1):
            wigner_D = (
                np.exp(1j * m * alpha_0)
                * np.exp(1j * mp * gamma_0)
                * lal.WignerdMatrix(l, m, mp, beta_0)
            )
            if (l, mp) not in hlm_L0:
                hlm_L0[(l, mp)] = wigner_D * hlm
            else:
                hlm_L0[(l, mp)] += wigner_D * hlm

    return hlm_L0



from dingo.gw.domains import UniformFrequencyDomain
from dingo.gw.waveform_generator.waveform_generator import WaveformGenerator, NewInterfaceWaveformGenerator
from dingo.gw.transforms.parameter_transforms import StandardizeParameters
import pytest
import numpy as np
import torch
import torch.distributions


@pytest.fixture
def uniform_fd_domain():
    p = {'f_min': 20.0, 'f_max': 4096.0, 'delta_f': 1.0/4.0}
    return UniformFrequencyDomain(**p)

@pytest.fixture
def aligned_spin_wf_parameters():
    parameters = {'chirp_mass': 34.0, 'mass_ratio': 0.35, 'chi_1': 0.2, 'chi_2': 0.1,
                  'theta_jn': 1.57, 'phase': 0.0, 'luminosity_distance': 1.0}
    f_ref = 20.0
    approximant = 'IMRPhenomPv2'
    return parameters, f_ref, approximant

@pytest.fixture
def precessing_spin_wf_parameters():
    parameters = {'chirp_mass': 34.0, 'mass_ratio': 0.35,
                  'a_1': 0.5, 'a_2': 0.2, 'tilt_1': 2*np.pi/3.0, 'tilt_2': np.pi/4.0,
                  'phi_12': np.pi/4.0, 'phi_jl': np.pi/3.0,
                  'theta_jn': 1.57, 'phase': 0.0, 'luminosity_distance': 1.0}
    f_ref = 100.0
    approximant = 'IMRPhenomPv2'
    return parameters, f_ref, approximant

@pytest.fixture(params=["aligned_spin_wf_parameters", "precessing_spin_wf_parameters"])
def wf_parameters(request):
    return request.getfixturevalue(request.param)


def test_waveform_generator_FD(uniform_fd_domain, wf_parameters):
    """Basic check that a waveform can be generated without error and
    it is consistent with the domain."""
    domain = uniform_fd_domain
    parameters, f_ref, approximant = wf_parameters

    wf_gen = WaveformGenerator(approximant, domain, f_ref)
    wf_dict = wf_gen.generate_hplus_hcross(parameters)

    assert len(wf_dict['h_plus']) == len(domain)
    assert domain()[domain.frequency_mask][0] == domain.f_min


def test_waveform_generator_FD_f_max_failure(precessing_spin_wf_parameters):
    """Specialized checks for time-domain waveforms.

    When f_max != 2**n, and a TD waveform model is being used the waveform
    can be generated, but the frequency spacing is then not consistent with
    between the value stored in the domain object and the value returned by LAL.

    In this case SimInspiralFD will return a **different delta_f** than used to
    define the domain due to automatically adjusting the sampling rate and this
    will trigger an exception in dingo's waveform_generator class.
    """
    # Common parameters
    parameters, f_ref, approximant = precessing_spin_wf_parameters
    approximant = 'SEOBNRv4PHM'


    # (1)
    # For lalsuite >= 7.11 when calling SimInspiralFD() f_max is rounded
    # to the next power-of-two multiple of deltaF -- see
    # https://git.ligo.org/lscsoft/lalsuite/-/commit/aaf02bc5aa9b13d32fccc077db6aa59e9bdaff4f
    # Check that generating a waveform with f_max not a power of two succeeds.
    # This includes a check in WaveformGenerator.generate_FD_waveform() which ensures
    # that the generated waveform agrees with the frequency grid defined in the domain.
    p_OK1 = {'f_min': 20.0, 'f_max': 896.0, 'delta_f': 1.0/8.0}
    domain_OK1 = UniformFrequencyDomain(**p_OK1)

    wf_gen = WaveformGenerator(approximant, domain_OK1, f_ref)
    wf_dict = wf_gen.generate_hplus_hcross(parameters)

    # (2)
    # Check that generating this waveform **succeeds** as expected.
    p_OK = {'f_min': 20.0, 'f_max': 1024.0, 'delta_f': 1.0/8.0}
    domain_OK = UniformFrequencyDomain(**p_OK)

    wf_gen = WaveformGenerator(approximant, domain_OK, f_ref)
    wf_dict = wf_gen.generate_hplus_hcross(parameters)


def test_standardize_parameters_on_distribution():
    """Check standardization of samples from a multi-normal distribution."""
    mean_ = torch.tensor([3.0, 2.0, 8.0])
    std_ = torch.tensor([2.0, 4.0, 7.0])
    n_samples = 100000
    parameters = torch.distributions.Normal(mean_, std_).sample((n_samples,)).numpy()
    samples = {'parameters': {'x': parameters}, 'waveform': None}
    tr = StandardizeParameters({'x': mean_.numpy()}, {'x': std_.numpy()})
    samples_tr = tr(samples)
    parameters_tr = samples_tr['parameters']['x']

    tol = 0.01
    assert np.all(np.abs(np.mean(parameters_tr, axis=0)) < tol)
    assert np.all(np.abs(np.std(parameters_tr, axis=0)) - np.ones(3) < tol)


def test_new_interface_extra_kwargs(uniform_fd_domain, precessing_spin_wf_parameters):
    """ When extra kwargs are provided at construction, they should be copied 
    through to the gwsignal parameter dictionary.
    """
    parameters, f_ref, _ = precessing_spin_wf_parameters

    wfg = NewInterfaceWaveformGenerator(
        approximant="SEOBNRv5PHM",
        domain=uniform_fd_domain,
        f_ref=f_ref,
        lmax_nyquist=3,
        postadiabatic=True,
        postadiabatic_type="analytic",
        enable_antisymmetric_modes=True,
        antisymmetric_modes_hm=True,
    )

    params_gwsignal = wfg._convert_parameters({**parameters, "f_ref": f_ref})

    assert params_gwsignal["lmax_nyquist"] == 3
    assert params_gwsignal["postadiabatic"] is True
    assert params_gwsignal["postadiabatic_type"] == "analytic"
    assert params_gwsignal["enable_antisymmetric_modes"] is True
    assert params_gwsignal["antisymmetric_modes_hm"] is True


import argparse

import torch

from dingo.core.utils.backward_compatibility import torch_load_with_fallback


def parse_args():

    parser = argparse.ArgumentParser()
    parser.add_argument("--checkpoint", type=str, required=True)
    parser.add_argument("--out_file", type=str, required=True)
    return parser.parse_args()


def main():
    args = parse_args()

    d, _ = torch_load_with_fallback(args.checkpoint)

    data_settings = d["metadata"]["train_settings"]["data"]
    f_min, f_max = data_settings["conditioning"]["frequency_range"]
    del data_settings["conditioning"]["frequency_range"]

    data_settings["domain_update"] = {"f_min": f_min, "f_max": f_max}

    torch.save(d, args.out_file)


if __name__ == "__main__":
    main()



DENSITY_RECOVERY_SETTINGS = {
    "ProxyRecoveryDefault": {
        "num_samples": 400_000,
        "threshold_std": 5,
        "nde_settings": {
            "model": {
                "posterior_model_type": "normalizing_flow",
                "posterior_kwargs": {
                    "num_flow_steps": 5,
                    "base_transform_kwargs": {
                        "hidden_dim": 256,
                        "num_transform_blocks": 4,
                        "activation": "elu",
                        "dropout_probability": 0.1,
                        "batch_norm": True,
                        "num_bins": 8,
                        "base_transform_type": "rq-coupling",
                    },
                },
            },
            "training": {
                "num_workers": 0,
                "train_fraction": 0.9,
                "batch_size": 4096,
                "epochs": 20,
                "optimizer": {
                    "type": "adam",
                    "lr": 0.002,
                },
                "scheduler": {
                    "type": "cosine",
                    "T_max": 20,
                },
            },
        },
    },
}

IMPORTANCE_SAMPLING_SETTINGS = {
    "PhaseRecoveryDefault": {
        "synthetic_phase": {
            "approximation_22_mode": False,
            "n_grid": 5001,
            "uniform_weight": 0.01,
        },
    },
    "MultibandingDefault": {
        "use_base_domain": True,
    },
}



from typing import Callable, Iterable

import numpy as np
import pandas as pd


def get_version():
    try:
        from dingo import __version__

        return __version__
    except ImportError:
        return None


def recursive_check_dicts_are_equal(dict_a, dict_b):
    if dict_a.keys() != dict_b.keys():
        return False
    else:
        for k, v_a in dict_a.items():
            v_b = dict_b[k]
            if type(v_a) != type(v_b):
                return False
            if type(v_a) == dict:
                if not recursive_check_dicts_are_equal(v_a, v_b):
                    return False
            elif not np.all(v_a == v_b):
                return False
    return True


def call_func_strict_output_dim(
    func: Callable[[int], Iterable[Iterable]],
    num_request: int,
    buffer_fraction: float = 0.02,
) -> Iterable[Iterable]:
    """
    Repeatedly calls a function until the output shape is the
    size of num_samples. This can be useful when a user requests
    N samples from a function, but because of failures in that
    function, only M < N sample are returned.

    Parameters
    ----------
    func : Callable[[int], Iterable[Iterable]]
        Function to repeatedly call. This should take an
        int and iterable of iterables. The int tells the
        function how many samples to generate.

        So for example if one passes 10 to the function,
        a possible output could be,
        [np.ndarray, pd.DataFrame, dict{"key":np.array}],
        where each entry has len of 10.
    num_request : int
        The output size of the function
    buffer_fraction : float, optional
        The fraction of extra samples to generate given
        we know the fraction of failed waveforms.

    Returns
    -------
    final_output: Iterable[Iterable]
        Iterable of iterables where each (sub) iterable
        has the len equal to num_request.
    """
    num_total, true_num_request = 0, num_request
    final_output = None
    num_times_zero_len_output = 0
    while num_total < true_num_request:
        output = func(num_request)

        # for the first iteration we don't need to append
        # to previous outputs
        if final_output is None:
            final_output = list(output)
            # estimating the failed fraction on the first iteration
            # is most accurate. This is because the first iteration
            # requests the largest number of samples
            failed_fraction = 1 - (len(output[0]) / num_request)
            if 1 - failed_fraction < 1e-6:
                raise ValueError(
                    f"""{failed_fraction * 100:1f}% of the function failed to generate.
                    Please check the inputs to {func} are sensible an try again. """
                )
        else:
            # appending the output to the final output
            for i, iterable in enumerate(output):
                if isinstance(iterable, np.ndarray):
                    final_output[i] = np.vstack([final_output[i], iterable])

                if isinstance(iterable, pd.DataFrame):
                    final_output[i] = pd.concat([final_output[i], iterable])

                if isinstance(iterable, dict):
                    for key, value in iterable.items():
                        final_output[i][key] = np.vstack([final_output[i][key], value])

        # making sure we don't enter an infinite loop, if the function
        # keeps failing to generate samples.
        if len(output[0]) == 0:
            num_times_zero_len_output += 1
            if num_times_zero_len_output > 5:
                raise ValueError(
                    """The function failed to generate any samples during the last 10 calls, check
                                 the inputs to the function are sensible and try again."""
                )

        # estimate the fraction of failures
        num_total += len(output[0])
        num_func_to_generate = num_request - num_total
        if failed_fraction == 0:
            break
        # generating extra calls given that we know the failing fraction
        num_request = int(
            (num_func_to_generate / failed_fraction) * (1 + buffer_fraction)
        )

    # truncating the final output to the correct size
    for i, iterable in enumerate(final_output):
        if isinstance(iterable, np.ndarray):
            final_output[i] = iterable[:true_num_request]

        if isinstance(iterable, pd.DataFrame):
            final_output[i] = iterable.iloc[:true_num_request]

        if isinstance(iterable, dict):
            for key, value in iterable.items():
                final_output[i][key] = value[:true_num_request]

    return final_output



import argparse
import json
from pathlib import Path
from pprint import pprint

import h5py
import torch
import yaml

from dingo.core.dataset import DingoDataset
from dingo.core.result import Result
from dingo.core.utils.backward_compatibility import torch_load_with_fallback
from dingo.gw.dataset import WaveformDataset
from dingo.gw.noise.asd_dataset import ASDDataset
from dingo.gw.SVD import SVDBasis


def ls():
    parser = argparse.ArgumentParser()
    parser.add_argument("file_name", type=str)
    args = parser.parse_args()

    path = Path(args.file_name)
    if path.suffix == ".pt":
        print("Extracting information about torch model.\n")
        d, _ = torch_load_with_fallback(path, preferred_map_location="meta")
        print(f"Version: {d.get('version')}\n")
        print(f"Model epoch: {d['epoch']}\n")
        print("Model metadata:")
        print(
            yaml.dump(
                d["metadata"],
                default_flow_style=False,
                sort_keys=False,
            )
        )

    elif path.suffix == ".hdf5":

        dataset_type = determine_dataset_type(args.file_name)

        if dataset_type == "gw_result" or dataset_type == "core_result":
            result = Result(file_name=args.file_name)
            print(f"Version: {result.version}")
            print("\nDingo Result\n" + "============\n")

            print(
                "Metadata\n"
                + "--------\n"
                + yaml.dump(
                    result.settings,
                    default_flow_style=False,
                    sort_keys=False,
                )
            )
            if result.event_metadata:
                print(
                    "Event information:\n"
                    + "------------------\n"
                    + yaml.dump(
                        result.event_metadata,
                        default_flow_style=False,
                        sort_keys=False,
                    ),
                )
            if result.importance_sampling_metadata is not None:
                print(
                    "Importance sampling:\n"
                    + "--------------------\n"
                    + yaml.dump(
                        result.importance_sampling_metadata,
                        default_flow_style=False,
                        sort_keys=False,
                    ),
                )
            if result.log_evidence:
                print("Summary:\n" + "--------")
                result.print_summary()

        elif dataset_type == "svd_basis":
            svd = SVDBasis(file_name=args.file_name)
            print(f"Dingo version: {svd.version}")
            print("\nSVD Basis\n" + "=========\n")

            print(f"Basis size: {svd.n}.")
            print("\nValidation summary:\n" + "-------------------")
            svd.print_validation_summary()

        elif dataset_type == "waveform_dataset":
            waveform_dataset = WaveformDataset(
                file_name=args.file_name, leave_waveforms_on_disk=True
            )
            print(f"Dingo version: {waveform_dataset.version}")
            print("\nWaveform dataset\n" + "================\n")

            print(f"Dataset size: {len(waveform_dataset)}")

            print(
                "\nSettings\n"
                + "--------\n"
                + yaml.dump(
                    waveform_dataset.settings,
                    default_flow_style=False,
                    sort_keys=False,
                )
            )

            if waveform_dataset.svd:
                svd = SVDBasis(dictionary=waveform_dataset.svd)
                print("\nSVD validation summary:\n" + "---------------------------")
                svd.print_validation_summary()

        elif dataset_type == "asd_dataset":
            asd_dataset = ASDDataset(file_name=args.file_name)
            print(f"Dingo version: {asd_dataset.version}")
            print("\nASD dataset\n" + "================\n")

            print(f"Dataset size: {asd_dataset.length_info}\n")
            print(f"GPS times (min/max): {asd_dataset.gps_info}")

            print(
                "\nSettings\n"
                + "--------\n"
                + yaml.dump(
                    asd_dataset.settings,
                    default_flow_style=False,
                    sort_keys=False,
                )
            )

        elif dataset_type == "trained_model":
            with h5py.File(args.file_name, "r") as f:
                print("Extracting information about torch model.\n")
                print(f"Version: {f.attrs['version']}")
                print(f"Model epoch: {f.attrs['epoch']}")
                print("Model metadata:")

                for d in ["model_kwargs", "metadata"]:
                    json_data = json.loads(f["serialized_dicts"][d][()])
                    print(f"\n{d}:\n" + "-" * (len(d) + 1))
                    pprint(json_data)

        else:
            # Legacy (before dataset_type identifier).
            try:
                svd = SVDBasis(file_name=args.file_name)
                print(f"SVD dataset of size n={svd.n}.")
                print("Validation summary:")
                svd.print_validation_summary()

            except KeyError:

                dataset = DingoDataset(
                    file_name=args.file_name,
                    data_keys=[
                        "svd",
                    ],
                )
                if dataset.settings is not None:
                    print(
                        yaml.dump(
                            dataset.settings,
                            default_flow_style=False,
                            sort_keys=False,
                        )
                    )
                if dataset.svd is not None:
                    svd = SVDBasis(dictionary=dataset.svd)
                    print("SVD validation summary:")
                    svd.print_validation_summary()

    elif path.suffix == ".yaml":
        with open(path, "r") as f:
            settings = yaml.safe_load(f)
        print(
            yaml.dump(
                settings,
                default_flow_style=False,
                sort_keys=False,
            )
        )

    else:
        print("File type unrecognized.")


def determine_dataset_type(file_name):
    with h5py.File(file_name, "r") as f:
        if "dataset_type" in f.attrs:
            return f.attrs["dataset_type"]
        else:
            return None



import torch
import numpy as np
from abc import abstractmethod

from torchdiffeq import odeint
from glasflow.nflows.utils.torchutils import repeat_rows, split_leading_dim

from .base_model import BasePosteriorModel

from dingo.core.nn.cfnets import create_cf


class ContinuousFlowPosteriorModel(BasePosteriorModel):
    """
    Class for posterior models based on continuous normalizing flows (CNFs).

    CNFs are parameterized by a vector field v(theta_t, t), that transports a simple
    base distribution (typically a gaussian N(0,1) with same dimension as theta) at
    time t=0 to the target distribution at time t=1. This vector field defines the flow
    via the ODE

                    d/dt f(theta, t) = v(f(theta, t), t).

    The vector field v is parameterized with a neural network. It is impractical to train
    this neural network (and thereby the CNF) directly with log-likelihood maximization,
    as solving the full ODE for each training iteration, requires thousands of
    vector field evaluations.

    Several alternative methods have been developed to make training CNFs more
    efficient. These directly regress on the vector field v (or a scaled version of v,
    such as the score). It has been shown that this can be done on a per-sample basis
    by adding noise to the parameters at various scales t. Specifically, a parameter
    sample theta is transformed as follows.

        t               ~ U[0, 1-eps)                               noise level
        theta_0         ~ N(0, 1)                                   sampled noise
        theta_1         = theta                                     pure sample
        theta_t         = c1(t) * theta_1 + c0(t) * theta_0         noisy sample

    Within that framework, one can employ different methods to learn the vector field v,
    such as flow matching or score matching. These have slightly different coefficients
    c1(t), c2(t) and training objectives.

    This class is intended to construct and hold a neural network for estimating the
    posterior density, as well as saving / loading, and training. It also has
    functionality for sampling and density evaluation.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.eps = 0
        self.time_prior_exponent = self.model_kwargs["posterior_kwargs"][
            "time_prior_exponent"
        ]
        self.theta_dim = self.metadata["train_settings"]["model"]["posterior_kwargs"][
            "input_dim"
        ]

    def sample_t(self, batch_size):
        t = (1 - self.eps) * torch.rand(batch_size, device=self.device)
        return t.pow(1 / (1 + self.time_prior_exponent))

    def sample_theta_0(self, batch_size):
        """Sample theta_0 from the gaussian prior."""
        return torch.randn(batch_size, self.theta_dim, device=self.device)

    @abstractmethod
    def evaluate_vector_field(self, t, theta_t, *context_data):
        """
        Evaluate the vector field v(t, theta_t, context_data) that generates the flow
        via the ODE

            d/dt f(theta_t, t, context) = v(f(theta_t, t, context), t, context).

        Parameters
        ----------
        t: float
            time (noise level)
        theta_t: torch.Tensor
            noisy parameters, perturbed with noise level t
        *context_data: list[torch.tensor]
            list with context data (GW data)
        """
        pass

    def rhs_of_joint_ode(self, t, theta_and_div_t, *context_data, hutchinson=False):
        """
        Returns the right hand side of the neural ODE that is used to evaluate the
        log_prob of theta samples. This is a joint ODE over the vector field and the
        divergence. By integrating this ODE, one can simultaneously trace the parameter
        sample theta_t and integrate the divergence contribution to the log_prob,
        see e.g., https://arxiv.org/abs/1806.07366 or Appendix C in
        https://arxiv.org/abs/2210.02747.

        Parameters
        ----------
        t: float
            time (noise level)
        theta_and_div_t: torch.Tensor
            concatenated tensor of (theta_t, div).
            theta_t: noisy parameters, perturbed with noise level t
        *context_data: list[torch.tensor]
            list with context data (GW data)

        Returns
        -------
        torch.Tensor
            vector field that generates the flow and its divergence (required for
            likelihood evaluation).
        """
        divergence_fun = compute_divergence
        if hutchinson:
            divergence_fun = compute_hutchinson_divergence

        theta_t = theta_and_div_t[:, :-1].detach()  # extract theta_t
        # context_data = context_data[0].detach().clone()
        t = t.detach()
        with torch.enable_grad():
            theta_t.requires_grad_(True)
            vf = self.evaluate_vector_field(t, theta_t, *context_data)
            div_vf = divergence_fun(vf, theta_t)
        return torch.cat((vf, -div_vf), dim=1)

    def initialize_network(self):
        model_kwargs = {
            k: v for k, v in self.model_kwargs.items() if k != "posterior_model_type"
        }
        if self.initial_weights is not None:
            model_kwargs["initial_weights"] = self.initial_weights
        self.network = create_cf(**model_kwargs)

    def sample(self, *context: torch.Tensor, num_samples: int = None):
        """
        Sample parameters theta from the posterior model,

        theta ~ p(theta | context)

        by solving an ODE forward in time.

        Parameters
        ----------
        context: torch.Tensor
            Context information (typically observed data). Should have a batch
            dimension (even if size B = 1).
        num_samples: int = 1
            Number of samples to generate.

        Returns
        -------
        samples: torch.Tensor
            Shape (B, num_samples, dim(theta))
        """
        context_size = context[0].shape[0]
        theta_0 = self.sample_theta_0(num_samples * context_size)
        context = [repeat_rows(c, num_reps=num_samples) for c in context]

        _, theta_1 = odeint(
            lambda t, theta_t: self.evaluate_vector_field(t, theta_t, *context),
            theta_0,
            self.integration_range,
            atol=1e-7,
            rtol=1e-7,
            method="dopri5",
        )

        theta_1 = split_leading_dim(theta_1, shape=[context_size, num_samples])

        self.network.train()
        return theta_1

    # MD: rename log_prob_batch, extract eps from self.epsilon_ode_integration
    def log_prob(self, theta: torch.Tensor, *context: torch.Tensor, hutchinson=False):
        """
        Evaluate the log posterior density,

        log p(theta | context)

        For this we solve an ODE backwards in time until we reach the initial pure
        noise distribution.

        There are two contributions, the log_prob of theta_0 (which is uniquely
        determined by theta) under the base distribution, and the integrated divergence
        of the vector field.

        Parameters
        ----------
        theta: torch.Tensor
            Parameter values at which to evaluate the density. Should have a batch
            dimension (even if size B = 1).
        context: torch.Tensor
            Context information (typically observed data). Must have context.shape[0] = B.
        hutchinson

        Returns
        -------
        log_prob: torch.Tensor
            Shape (B,)
        """
        theta_and_div_init = torch.cat(
            (theta, torch.zeros((theta.shape[0],), device=theta.device).unsqueeze(1)),
            dim=1,
        )
        _, theta_and_div_0 = odeint(
            lambda t, theta_and_div_t: self.rhs_of_joint_ode(
                t, theta_and_div_t, *context, hutchinson=hutchinson
            ),
            theta_and_div_init,
            torch.flip(
                self.integration_range, dims=(0,)
            ),  # integrate backwards in time, [1-eps, 0]
            atol=1e-7,
            rtol=1e-7,
            options={"norm": norm_without_divergence_component},
            method="dopri5",
        )
        # return log p_0(x_0) - divergence|0
        theta_0, divergence = theta_and_div_0[:, :-1], theta_and_div_0[:, -1]
        log_prior = compute_log_prior(theta_0)
        return (log_prior - divergence).detach()

    def sample_and_log_prob(self, *context: torch.Tensor, num_samples: int = None):
        """
        Sample parameters theta from the posterior model,

        theta ~ p(theta | context)

        and also return the log_prob. This is more efficient than calling sample_batch
        and log_prob_batch separately.

        If d/dt [phi(t), f(t)] = rhs joint with initial conditions [theta_0,
        log p(theta_0)], where theta_0 ~ p_0(theta_0), then [phi(1), f(1)] = [theta_1,
        log p(theta_0) + log p_1(theta_1) - log p(theta_0)] = [theta_1, log p_1(theta_1)].

        Parameters
        ----------
        context: torch.Tensor
            Context information (typically observed data). Should have a batch
            dimension (even if size B = 1).
        num_samples: int = 1
            Number of samples to generate.

        Returns
        -------
        samples, log_prob: torch.Tensor, torch.Tensor
            Shapes (B, num_samples, dim(theta)), (B, num_samples)
        """

        context_size = context[0].shape[0]
        theta_0 = self.sample_theta_0(num_samples * context_size)
        context = [repeat_rows(c, num_reps=num_samples) for c in context]

        log_prior = compute_log_prior(theta_0)
        theta_and_div_init = torch.cat((theta_0, log_prior.unsqueeze(1)), dim=1)

        _, theta_and_div_1 = odeint(
            lambda t, theta_and_div_t: self.rhs_of_joint_ode(
                t, theta_and_div_t, *context
            ),
            theta_and_div_init,
            self.integration_range,  # integrate forwards in time, [0, 1-eps]
            atol=1e-7,
            rtol=1e-7,
            method="dopri5",
            options={"norm": norm_without_divergence_component},
        )

        theta_1, log_prob_1 = theta_and_div_1[:, :-1], theta_and_div_1[:, -1]

        theta_1 = split_leading_dim(theta_1, shape=[context_size, num_samples])
        log_prob_1 = split_leading_dim(log_prob_1, shape=[context_size, num_samples])

        return theta_1, log_prob_1

    @property
    def integration_range(self):
        """
        Integration range for ODE. We integrate in the range [0, 1-self.eps]. For score
        matching, self.eps > 0 is required for stability. For flow matching we can have
        self.eps = 0.
        """
        return torch.tensor([0.0, 1.0 - self.eps]).type(torch.float32).to(self.device)


def compute_divergence(y, x):
    div = 0.0
    with torch.enable_grad():
        y.requires_grad_(True)
        x.requires_grad_(True)
        for i in range(y.shape[-1]):
            div += torch.autograd.grad(
                y[..., i], x, torch.ones_like(y[..., i]), retain_graph=True
            )[0][..., i : i + 1]
        return div


def compute_hutchinson_divergence(y, x):
    with torch.enable_grad():
        x.requires_grad_(True)
        eps = torch.randn_like(y)
        y_cdot_eps = torch.sum(eps * y)
        x.retain_grad()
        y_cdot_eps.backward()
        return torch.sum(eps * x.grad, dim=-1, keepdim=True)


def norm_without_divergence_component(y):
    return torch.linalg.norm(y[:, :-1])


def compute_log_prior(theta_0):
    N = theta_0.shape[1]
    return -N / 2.0 * np.log(2 * np.pi) - torch.sum(theta_0**2, dim=1) / 2.0



from typing import Optional

import numpy as np

from dingo.gw.inference.gw_samplers import GWSampler


def prepare_log_prob(
    sampler,
    num_samples: int,
    nde_settings: dict,
    batch_size: Optional[int] = None,
    threshold_std: Optional[float] = np.inf,
    remove_init_outliers: Optional[float] = 0.0,
    low_latency_label: str = None,
    outdir: str = None,
):
    """
    Prepare gnpe sampling with log_prob. This is required, since in its vanilla
    form gnpe does not provide the density for its samples.

    Specifically, we train an unconditional neural density estimator (nde) for the
    gnpe proxies. This requires running the gnpe sampler till convergence, and
    extracting the gnpe proxies after the final gnpe iteration. The nde is trained
    to match the distribution over gnpe proxies, which provides a way of rapidly
    sampling (converged!) gnpe proxies *and* evaluating the log_prob.

    After this preparation step, self.run_sampler can leverage
    self.gnpe_proxy_sampler (which is based on the aforementioned trained nde) to
    sample gnpe proxies, such that one gnpe iteration is sufficient. The
    log_prob of
    the samples in the *joint* space (inference parameters + gnpe proxies) is then
    simply given by the sum of the corresponding log_probs (from self.model and
    self.gnpe_proxy_sampler.model).

    Parameters
    ----------
    num_samples: int
        number of samples for training of nde
    batch_size: int = None
        batch size for sampler
    threshold_std: float = np.inf
        gnpe proxies deviating by more then threshold_std standard deviations from
        the proxy mean (along any axis) are discarded.
    low_latency_label: str = None
        File label for low latency samples (= samples used for training nde).
        If None, these samples are not saved.
    outdir: str = None
        Directory in which low latency samples are saved. Needs to be set if
        low_latency_label is not None.
    """
    sampler.remove_init_outliers = remove_init_outliers
    sampler.run_sampler(num_samples, batch_size)
    if low_latency_label is not None:
        sampler.to_hdf5(label=low_latency_label, outdir=outdir)
    result = sampler.to_result()
    nde_settings["training"]["device"] = str(sampler.model.device)
    unconditional_model = result.train_unconditional_flow(
        sampler.gnpe_proxy_parameters,
        nde_settings,
        threshold_std=threshold_std,
    )

    # Prepare sampler with unconditional model as initialization. This should only use
    # one iteration and also not remove any outliers.
    sampler.init_sampler = GWSampler(model=unconditional_model)
    sampler.num_iterations = 1
    sampler.remove_init_outliers = 0.0  # Turn off for final sampler.



import functools

import numpy as np

import bilby
import bilby_pipe.input as input_mod
import bilby_pipe.create_injections as ci

_bilby_pipe_create_injection_file = ci.create_injection_file


@functools.wraps(_bilby_pipe_create_injection_file)
def create_injection_file(*args, **kwargs):
    # Monkey-patch create_injection_file() to allow for an offset in the time. This is
    # needed if the DINGO network time prior is not centered at 0.0, e.g., if we have a
    # Dirac delta prior.

    Toffset = kwargs.pop("Toffset", 0.0)

    if kwargs.get("trigger_time") is not None:
        kwargs["trigger_time"] += Toffset
    if kwargs.get("gpstimes") is not None:
        gpstimes = kwargs["gpstimes"]
        kwargs["gpstimes"] = (
            (np.asarray(gpstimes) + Toffset).tolist()
            if isinstance(gpstimes, list)
            else gpstimes + Toffset
        )

    return _bilby_pipe_create_injection_file(*args, **kwargs)


create_injection_file.__doc__ = (
    "Add a `Toffset` keyword to offset the trigger and GPS times before creating injections.\n\n"
    + _bilby_pipe_create_injection_file.__doc__
)

ci.create_injection_file = create_injection_file

_bilby_pipe_get_time_prior = input_mod.get_time_prior


@functools.wraps(_bilby_pipe_get_time_prior)
def get_time_prior(time, uncertainty, name="geocent_time", latex_label="$t_c$"):
    # Monkey-patch to allow for DeltaFunction priors.
    if uncertainty == 0.0:
        return bilby.core.prior.DeltaFunction(
            peak=time,
            name=name,
            latex_label=latex_label,
            unit="$s$",
        )
    elif uncertainty > 0.0:
        return _bilby_pipe_get_time_prior(
            time, uncertainty, name=name, latex_label=latex_label
        )
    else:
        raise ValueError(f"Time uncertainty {uncertainty} < 0.0.")


get_time_prior.__doc__ = (
    "A bilby.core.prior.DeltaFunction for the time parameter, if uncertainty == 0.0.\n\n"
    + _bilby_pipe_get_time_prior.__doc__
)


input_mod.get_time_prior = get_time_prior



from typing import Optional
import numpy as np

from dingo.gw.domains import MultibandedFrequencyDomain, UniformFrequencyDomain


class DecimateAll(object):
    """Transform operator for decimation to multibanded frequency domain."""

    def __init__(
        self,
        multibanded_frequency_domain: MultibandedFrequencyDomain,
    ):
        """
        Parameters
        ----------
        multibanded_frequency_domain: MultibandedFrequencyDomain
            New domain of the decimated data. Original data must be in
            multibanded_frequency_domain.base_domain
        """
        self.multibanded_frequency_domain = multibanded_frequency_domain

    def __call__(self, input_sample: dict) -> dict:
        """
        Parameters
        ----------
        input_sample : dict
            For each decimation_key in self.decimation_keys, sample[decimation_key]
            should be (1) a dict with arrays containing data to be transformed,
            or (2) an array with data to be transformed.

        Returns
        -------
        dict of the same form as the input, but with transformed (decimated) data.
        """
        sample = input_sample.copy()
        decimate_recursive(sample, self.multibanded_frequency_domain)
        return sample


def decimate_recursive(d: dict, mfd: MultibandedFrequencyDomain):
    """
    In-place decimation of nested dicts of arrays.

    Parameters
    ----------
    d : dict
        Nested dictionary to decimate.
    mfd : MultibandedFrequencyDomain
    """
    for k, v in d.items():
        if isinstance(v, dict):
            decimate_recursive(v, mfd)
        elif isinstance(v, np.ndarray):
            if v.shape[-1] == len(mfd.base_domain):
                d[k] = mfd.decimate(v)
        else:
            raise ValueError(f"Cannot decimate item of type {type(v)}.")


class DecimateWaveformsAndASDS(object):
    """Transform operator for decimation of unwhitened waveforms and corresponding ASDS
    to multibanded frequency domain (MFD).


    For decimation, we have two options.


    1) decimation_mode = whitened
    In this case, the GW data is whitened first,

            dw = d / ASD,

    and then decimated to the MFD,

            dw_mfd = decimate(dw).

    In this case, the effective ASD in the MFD is given by

            ASD_mfd = 1 / decimate(1 / ASD).

    See [1] for details. ASD_mfd can then be provided to the inference network.


    2) decimation_mode = unwhitened
    In this case, the GW data is first decimated,

            d_mfd = decimate(d)

    and then whitened.

            dw_mfd = d_mfd / ASD_mfd.

    In this case, the ASD_mfd that whitens the data d_mfd is given by

            ASD_mfd = decimate(ASD ** 2) ** 0.5.

    In other words, in this case we need to decimate the *PSD*. See [1] for details.


    Method 1) better preserves the signal.

    [1] https://github.com/dingo-gw/dingo/blob/fede5c01524f3e205acf5750c0a0f101ff17e331/binary_neutron_stars/prototyping/psd_decimation.ipynb
    """

    def __init__(
        self,
        multibanded_frequency_domain: MultibandedFrequencyDomain,
        decimation_mode: str,
    ):
        """
        Parameters
        ----------
        multibanded_frequency_domain: MultibandedFrequencyDomain
            New domain of the decimated data. Original data must be in
            multibanded_frequency_domain.base_domain
        decimation_mode: str
            One of ["whitened", "unwhitened"]. Determines whether decimation is
            performed on whitened data or on unwhitened data. See class docstring for
            details.
        """
        self.multibanded_frequency_domain = multibanded_frequency_domain
        if decimation_mode not in ["whitened", "unwhitened"]:
            raise ValueError(
                f"Unsupported decimation mode {decimation_mode}, needs to be one of "
                f'["whitened", "unwhitened"].'
            )
        self.decimation_mode = decimation_mode

    def __call__(self, input_sample: dict) -> dict:
        """
        Parameters
        ----------
        input_sample : dict
            Values of sample["waveform"] should be arrays containing waveforms to be
            transformed, Values of sample["asds"] should be arrays containing asds
            to be transformed.

        Returns
        -------
        dict of the same form as the input, but with transformed (decimated) waveforms
        and asds.
        """
        sample = input_sample.copy()

        # Only decimate the data if it is in the base domain. If it has already been
        # decimated, do not change it.

        if check_sample_in_domain(
            sample, self.multibanded_frequency_domain.base_domain
        ):
            if self.decimation_mode == "whitened":
                whitened_waveforms = {
                    k: v / sample["asds"][k] for k, v in sample["waveform"].items()
                }
                whitened_waveforms_dec = {
                    k: self.multibanded_frequency_domain.decimate(v)
                    for k, v in whitened_waveforms.items()
                }
                asds_dec = {
                    k: 1 / self.multibanded_frequency_domain.decimate(1 / v)
                    for k, v in sample["asds"].items()
                }
                # color the whitened waveforms with the effective asd
                waveform_dec = {
                    k: v * asds_dec[k] for k, v in whitened_waveforms_dec.items()
                }
                sample["waveform"] = waveform_dec
                sample["asds"] = asds_dec

            elif self.decimation_mode == "unwhitened":
                sample["waveform"] = {
                    k: self.multibanded_frequency_domain.decimate(v)
                    for k, v in sample["waveform"].items()
                }
                sample["asds"] = {
                    k: self.multibanded_frequency_domain.decimate(v**2) ** 0.5
                    for k, v in sample["asds"].items()
                }

            else:
                raise NotImplementedError()

        return sample


def check_sample_in_domain(sample, domain: UniformFrequencyDomain) -> bool:
    lengths = []
    base_domain_length = len(domain)
    for k in ["waveform", "asds"]:
        lengths += [d.shape[-1] for d in sample[k].values()]
    if all(l == base_domain_length for l in lengths):
        return True
    else:
        return False


class CropMaskStrainRandom(object):
    """Apply random cropping of strain, by masking waveform and ASD outside the crop."""

    def __init__(
        self,
        domain: UniformFrequencyDomain | MultibandedFrequencyDomain,
        f_min_upper: Optional[float] = None,
        f_max_lower: Optional[float] = None,
        cropping_probability: float = 1.0,
        independent_detectors: bool = True,
        independent_lower_upper: bool = True,
        deterministic_fmin_fmax: Optional[
            list[float | None] | list[list[float | None]]
        ] = None,
    ):
        """
        Parameters
        ----------
        domain: UniformFrequencyDomain | MultibandedFrequencyDomain
            Domain of the waveform data, has to be a frequency domain type.
        f_min_upper: float
            New f_min is sampled in range [domain.f_min, f_min_upper].
            Sampling of f_min is uniform in bins (not in frequency) when the frequency
            domain is not uniform (e.g., MultibandedFrequencyDomain).
        f_max_lower: float
            New f_max is sampled in range [f_max_lower, domain.f_max].
            Sampling of f_max is uniform in bins (not in frequency) when the frequency
            domain is not uniform (e.g., MultibandedFrequencyDomain).
        cropping_probability: float
            probability for a given sample to be cropped
        independent_detectors: bool
            If True, crop boundaries are sampled independently for different detectors.
        independent_lower_upper: bool
            If True, the cropping probability is applied to lower and upper boundaries
            individually. If False, then with a probability of P = cropping_probability
            both lower and upper cropping is applied, and with 1-P, no cropping is
            applied from either direction.
        deterministic_fmin_fmax: None | list[float | None] | list[list[float | None]]
            If not None, then cropping will be applied with fixed f_min and f_max,
            as specified by this argument. Can either be a single list [f_min, f_max],
            or a list of lists [[f_min_1, f_max_1], [f_min_2, f_max_2], ...] for the
            different detectors.
        """
        self._check_inputs(
            domain,
            f_min_upper,
            f_max_lower,
            cropping_probability,
            deterministic_fmin_fmax,
        )
        self._deterministic_fmin_fmax = deterministic_fmin_fmax
        self.cropping_probability = cropping_probability
        self.independent_detectors = independent_detectors
        self.independent_lower_upper = independent_lower_upper
        self.frequencies = domain()[domain.min_idx :]
        self.len_domain = len(self.frequencies)

        # for stochastic sampling of cropping boundaries
        self._idx_bound_f_max = self._get_domain_idx(f_max_lower, self.len_domain - 1)
        self._idx_bound_f_min = self._get_domain_idx(f_min_upper, 0)

        # for deterministic cropping boundaries
        self._deterministic_mask = None
        if deterministic_fmin_fmax is not None:
            self._initialize_deterministic_mask(deterministic_fmin_fmax)

    def _sample_upper_bound_indices(self, shape: list[int]) -> np.ndarray:
        """Sample indices for upper crop boundaries."""
        assert self._deterministic_mask is None
        return np.random.randint(self._idx_bound_f_max, self.len_domain, shape)

    def _sample_lower_bound_indices(self, shape: list[int]) -> np.ndarray:
        """Sample indices for lower crop boundaries."""
        assert self._deterministic_mask is None
        return np.random.randint(0, self._idx_bound_f_min + 1, shape)

    def _get_domain_idx(self, f_value: float, default_value: int = 0) -> int:
        if f_value is not None:
            return np.argmin(np.abs(f_value - self.frequencies)).item()
        else:
            return default_value

    def _initialize_deterministic_mask(self, deterministic_fmin_fmax):
        if not isinstance(deterministic_fmin_fmax[0], list):
            deterministic_fmin_fmax = [deterministic_fmin_fmax]
        masking_indices = []
        for f_min, f_max in deterministic_fmin_fmax:
            self._check_fmin_fmax(f_min, f_max)
            idx_min = self._get_domain_idx(f_min, 0)
            idx_max = self._get_domain_idx(f_max, self.len_domain - 1)
            if idx_min >= idx_max:
                raise ValueError(f"f_min ({f_min}) must be less than f_max ({f_max}).")
            masking_indices.append([idx_min, idx_max])
        masking_indices = np.array(masking_indices)
        lower = masking_indices[:, 0]
        upper = masking_indices[:, 1]
        mask_lower = np.arange(len(self.frequencies))[None, :] >= lower[:, None]
        mask_upper = np.arange(len(self.frequencies))[None, :] <= upper[:, None]
        # combine masks
        mask = mask_lower * mask_upper
        # broadcast (detectors, freq) => (batch_size, detectors, channels, freq)
        self._deterministic_mask = mask[None, :, None, :]

    def _check_fmin_fmax(self, f_min, f_max):
        """Check that domain.f_min < fmin < f_max < domain.f_max."""
        domain_f_min = self.frequencies[0]
        domain_f_max = self.frequencies[-1]
        if f_min is not None and f_max is not None:
            if not (domain_f_min < f_min < f_max < domain_f_max):
                raise ValueError(
                    f"Expected {f_min} in range ({domain_f_min}, {domain_f_max}) "
                    f"and {f_max} in range ({domain_f_min}, {domain_f_max})."
                )
        elif f_min is not None:
            if not (domain_f_min < f_min < domain_f_max):
                raise ValueError(
                    f"Expected {f_min} in range ({domain_f_min}, {domain_f_max})."
                )
        elif f_max is not None:
            if not (domain_f_min < f_max < domain_f_max):
                raise ValueError(
                    f"Expected {f_max} in range ({domain_f_min}, {domain_f_max})."
                )

    def _check_inputs(
        self,
        domain: UniformFrequencyDomain | MultibandedFrequencyDomain,
        f_min_upper: float,
        f_max_lower: float,
        cropping_probability: float,
        deterministic_fmin_fmax: Optional[
            list[float | None] | list[list[float | None]]
        ],
    ):

        # check domain
        if not isinstance(domain, (UniformFrequencyDomain, MultibandedFrequencyDomain)):
            raise ValueError(
                f"Domain should be a frequency domain type, got {type(domain)}."
            )
        self.frequencies = domain()[domain.min_idx :]

        # check validity of ranges for non-deterministic case
        if deterministic_fmin_fmax is None:
            self._check_fmin_fmax(f_min_upper, f_max_lower)
            if not 0 <= cropping_probability <= 1.0:
                raise ValueError(
                    f"Cropping probability should be in [0, 1], got {cropping_probability}."
                )

        # check deterministic case
        else:
            # check that we don't have non-trivial f_min_upper, f_max_lower,
            # cropping_probability in deterministic case
            if f_min_upper is not None or f_max_lower is not None:
                raise ValueError(
                    "If deterministic_fmin_fmax is set, f_min_upper and f_max_lower should"
                    " not be set."
                )
            if cropping_probability < 1.0:
                raise ValueError(
                    f"cropping_probability must be 1.0 when deterministic = True, got "
                    f"{cropping_probability}."
                )

    def __call__(self, input_sample: dict) -> dict:
        """
        Parameters
        ----------
        input_sample : dict
            sample["waveform"]: Dict with values being arrays containing waveforms,
            or torch Tensor with the waveform.

        Returns
        -------
        dict of the same form as the input, but with transformed (crop-masked) waveforms.
        """
        sample = input_sample.copy()
        strain = sample["waveform"]
        if strain.shape[-1] != self.len_domain:
            raise ValueError(
                f"Expected waveform input of shape [..., {self.len_domain}], "
                f"got {strain.shape}."
            )

        if self._deterministic_mask is not None:
            if self._deterministic_mask.shape[1] not in [1, strain.shape[1]]:
                raise ValueError(
                    f"Deterministic masking range needs to be either supplied for all "
                    f"detectors, or for each individual one. Strain provided in "
                    f"{strain.shape[1]} detectors, so deterministic range should be "
                    f"be provided as a single [fmin, fmax] pair or as {strain.shape[1]} "
                    f"pairs, got {self._deterministic_mask.shape[1]}."
                )
            strain = np.where(self._deterministic_mask, strain, 0)
            sample["waveform"] = strain
            return sample

        # Sample boundary indices for crops
        constant_ax = 3 - self.independent_detectors
        lower = self._sample_lower_bound_indices(strain.shape[:-constant_ax])
        upper = self._sample_upper_bound_indices(strain.shape[:-constant_ax])

        # Only apply crops to a fraction of self.cropping_probability
        if self.cropping_probability < 1:
            mask = np.random.uniform(size=lower.shape) <= self.cropping_probability
            lower = np.where(mask, lower, 0)
            if self.independent_lower_upper:
                mask = np.random.uniform(size=lower.shape) <= self.cropping_probability
            upper = np.where(mask, upper, self.len_domain)

        # Broadcast boundaries and apply cropping
        mask_lower = np.arange(self.len_domain) >= lower[(...,) + (None,) * constant_ax]
        mask_upper = np.arange(self.len_domain) <= upper[(...,) + (None,) * constant_ax]
        strain = np.where(mask_lower, strain, 0)
        strain = np.where(mask_upper, strain, 0)
        sample["waveform"] = strain

        return sample


class MaskDataForFrequencyRangeUpdate(object):
    """
    Set waveform to zero and ASD to one according to minimum_frequency and maximum_frequency.
    """

    def __init__(
        self,
        domain: UniformFrequencyDomain | MultibandedFrequencyDomain,
        minimum_frequency: Optional[float | dict[str, float]] = None,
        maximum_frequency: Optional[float | dict[str, float]] = None,
        print_output: bool = False,
    ):
        """
        Parameters
        ----------
        minimum_frequency: Optional[float | dict[str, float]]
            Update of f_min. If a float, the same value will be used for all detectors.
        maximum_frequency: Optional[float | dict[str, float]]
            Update of f_max. If a float, the same value will be used for all detectors.
        print_output: bool
            Whether to write print statements to the console.
        """
        self.sample_frequencies = domain.sample_frequencies
        self.minimum_frequency = minimum_frequency
        self.maximum_frequency = maximum_frequency

        if print_output:
            print(
                f"Transform MaskDataForFrequencyRangeUpdate activated:"
                f"  Settings: \n"
                f"    - Minimum_frequency update: {self.minimum_frequency}\n"
                f"    - Maximum_frequency update: {self.maximum_frequency}\n"
            )

    def __call__(self, input_sample: dict) -> dict:
        """
        Parameters
        ----------
        input_sample: Dict
            Values for keys
            - 'waveform', dict containing waveform for each detector
            - 'asds', dict containing asds for each detector

        Returns
        ----------
        sample: Dict
            input_sample with modified value for key
            - 'waveform'
            - 'asds'

        """
        sample = input_sample.copy()

        ifos = list(sample["waveform"].keys())
        frequency_masks = create_mask_based_on_frequency_update(
            sample_frequencies=self.sample_frequencies,
            detectors=ifos,
            minimum_frequency=self.minimum_frequency,
            maximum_frequency=self.maximum_frequency,
        )

        # Update waveforms and ASDs based on masks. Be sure not to modify dicts or
        # arrays in-place.
        sample["waveform"] = {
            ifo: np.where(frequency_masks[ifo], sample["waveform"][ifo], 0.0)
            for ifo in ifos
        }
        sample["asds"] = {
            ifo: np.where(frequency_masks[ifo], sample["asds"][ifo], 1.0)
            for ifo in ifos
        }

        return sample


def create_mask_based_on_frequency_update(
    sample_frequencies: np.array,
    detectors: list[str],
    minimum_frequency: Optional[float | dict[str, float]] = None,
    maximum_frequency: Optional[float | dict[str, float]] = None,
) -> dict[str : np.array]:
    """
    Creates a mask for each detector containing True for sample_frequencies not affected by the frequency updates
    for minimum and maximum frequencies.

    Parameters
    ----------
    sample_frequencies: np.array
        Frequency values for which we want to create a mask. Assumed to be the same for all detectors.
    detectors: list[str]
        List of detector names.
    minimum_frequency: Optional[float | dict[str, float]]
        Update of f_min. If a float, the same value will be used for all detectors.
    maximum_frequency: Optional[float | dict[str, float]]
        Update of f_max. If a float, the same value will be used for all detectors.

    Returns
    ----------
    frequency_masks: dict
        Dictionary providing a boolean mask for the sample frequencies of each detector based on provided frequency
        updates. True for values to keep, False for values to mask.
    """
    frequency_masks = {
        i: np.ones_like(sample_frequencies, dtype=bool) for i in detectors
    }
    for d in detectors:
        # Update frequency_masks based on minimum_frequency
        if minimum_frequency is not None:
            # Same for all detectors
            if isinstance(minimum_frequency, float) or isinstance(
                minimum_frequency, int
            ):
                mask_min = sample_frequencies >= minimum_frequency
            # Different for each detector
            elif isinstance(minimum_frequency, dict):
                mask_min = sample_frequencies >= minimum_frequency[d]
            else:
                raise ValueError(
                    f"minimum-frequency has to be dict, int or float, not {type(minimum_frequency)}. "
                )
            frequency_masks[d] = frequency_masks[d] & mask_min

        # Update frequency_masks based on maximum_frequency
        if maximum_frequency is not None:
            # Same for all detectors
            if isinstance(maximum_frequency, float) or isinstance(
                maximum_frequency, int
            ):
                mask_max = sample_frequencies <= maximum_frequency
            # Different for each detector
            elif isinstance(maximum_frequency, dict):
                mask_max = sample_frequencies <= maximum_frequency[d]
            else:
                raise ValueError(
                    f"maximum-frequency has to be dict, int or float, not {type(maximum_frequency)}. "
                )
            frequency_masks[d] = frequency_masks[d] & mask_max

    return frequency_masks



import numpy as np
import pytest
from copy import deepcopy

from dingo.gw.domains import UniformFrequencyDomain, MultibandedFrequencyDomain
from dingo.gw.likelihood import StationaryGaussianGWLikelihood


@pytest.fixture
def ufd_setup():
    f_min = 20.0
    f_max = 1024.0
    T = 8.0
    domain = UniformFrequencyDomain(f_min, f_max, delta_f=1 / T)

    # Set waveform below f_min to 0. (This is important for the test to work!)
    waveform = {
        "H1": np.where(domain.frequency_mask, (1.0 + 1j) * 1e-20, 0.0),
        "L1": np.where(domain.frequency_mask, (1.0 + 1j) * 1e-20, 0.0),
        "V1": np.where(domain.frequency_mask, (1.0 + 1j) * 1e-20, 0.0),
    }
    # Set asds below f_min to 1.
    asds = {
        "H1": np.where(domain.frequency_mask, 1e-20, 1.0),
        "L1": np.where(domain.frequency_mask, 1e-20, 1.0),
        "V1": np.where(domain.frequency_mask, 1e-20, 1.0),
    }

    sample = {"waveform": waveform, "asds": asds}

    likelihood_kwargs = {
        "wfg_kwargs": {
            "approximant": "IMRPhenomXPHM",
            "f_ref": 20.0,
            "spin_conversion_phase": 0.0,
        },
        "wfg_domain": domain,
        "data_domain": domain,
        "event_data": sample,
        "t_ref": 1248242632.0,
    }
    # Create data with smaller range, e.g. [20., 512.]
    f_max_new = 512.0
    domain_small = UniformFrequencyDomain(
        f_min, f_max_new, delta_f=1 / T
    )

    waveform_small = {
        "H1": np.where(domain_small.frequency_mask, (1.0 + 1j) * 1e-20, 0.0),
        "L1": np.where(domain_small.frequency_mask, (1.0 + 1j) * 1e-20, 0.0),
        "V1": np.where(domain_small.frequency_mask, (1.0 + 1j) * 1e-20, 0.0),
    }
    asds_small = {
        "H1": np.where(domain_small.frequency_mask, 1e-20, 1.0),
        "L1": np.where(domain_small.frequency_mask, 1e-20, 1.0),
        "V1": np.where(domain_small.frequency_mask, 1e-20, 1.0),
    }

    sample_small = {"waveform": waveform_small, "asds": asds_small}
    likelihood_kwargs_small = {
        "wfg_kwargs": {
            "approximant": "IMRPhenomXPHM",
            "f_ref": 20.0,
            "spin_conversion_phase": 0.0,
        },
        "wfg_domain": domain_small,
        "data_domain": domain_small,
        "event_data": sample_small,
        "t_ref": 1248242632.0,
    }

    likelihood_kwargs_masked = deepcopy(likelihood_kwargs)
    likelihood_kwargs_masked["frequency_update"] = {
        "maximum_frequency": f_max_new,
    }

    theta = {
        "chirp_mass": 50.0,
        "mass_ratio": 0.8,
        "a_1": 0.3,
        "a_2": 0.4,
        "tilt_1": 1.3,
        "tilt_2": 1.4,
        "phi_12": 0.3,
        "phi_jl": 3.3,
        "theta_jn": 1.2,
        "luminosity_distance": 4000.0,
        "geocent_time": -0.03,
        "dec": -0.3,
        "ra": 1.5,
        "psi": 1.2,
        "phase": 1.4,
    }

    return likelihood_kwargs, likelihood_kwargs_small, likelihood_kwargs_masked, theta


@pytest.fixture
def mfd_setup():
    nodes = [20.0, 34.0, 46.0, 62.0, 78.0, 1038.0]
    f_min = 20.0
    f_max = 1038.0
    T = 8.0
    base_domain = UniformFrequencyDomain(
        f_min=f_min, f_max=f_max, delta_f=1 / T
    )
    domain = MultibandedFrequencyDomain(
        nodes=nodes, delta_f_initial=1 / T, base_domain=base_domain
    )

    # Set waveform below f_min to 0. (This is important for the test to work!)
    waveform = {
        "H1": np.where(domain.frequency_mask, (1.0 + 1j) * 1e-20, 0.0),
        "L1": np.where(domain.frequency_mask, (1.0 + 1j) * 1e-20, 0.0),
        "V1": np.where(domain.frequency_mask, (1.0 + 1j) * 1e-20, 0.0),
    }
    # Set asds below f_min to 1.
    asds = {
        "H1": np.where(domain.frequency_mask, 1e-20, 1.0),
        "L1": np.where(domain.frequency_mask, 1e-20, 1.0),
        "V1": np.where(domain.frequency_mask, 1e-20, 1.0),
    }

    sample = {"waveform": waveform, "asds": asds}

    likelihood_kwargs = {
        "wfg_kwargs": {
            "approximant": "IMRPhenomXPHM",
            "f_ref": 20.0,
            "spin_conversion_phase": 0.0,
        },
        "wfg_domain": domain,
        "data_domain": domain,
        "event_data": sample,
        "t_ref": 1248242632.0,
    }
    # Create data with smaller range, e.g. [20., 512.]
    f_max_new = 512.0
    nodes_small = [20.0, 34.0, 46.0, 62.0, 78.0, f_max_new]
    base_domain_small = UniformFrequencyDomain(
        f_min, f_max_new, delta_f=1 / T
    )
    domain_small = MultibandedFrequencyDomain(
        nodes=nodes_small, delta_f_initial=1 / T, base_domain=base_domain_small
    )

    waveform_small = {
        "H1": np.where(domain_small.frequency_mask, (1.0 + 1j) * 1e-20, 0.0),
        "L1": np.where(domain_small.frequency_mask, (1.0 + 1j) * 1e-20, 0.0),
        "V1": np.where(domain_small.frequency_mask, (1.0 + 1j) * 1e-20, 0.0),
    }
    asds_small = {
        "H1": np.where(domain_small.frequency_mask, 1e-20, 1.0),
        "L1": np.where(domain_small.frequency_mask, 1e-20, 1.0),
        "V1": np.where(domain_small.frequency_mask, 1e-20, 1.0),
    }

    sample_small = {"waveform": waveform_small, "asds": asds_small}
    likelihood_kwargs_small = {
        "wfg_kwargs": {
            "approximant": "IMRPhenomXPHM",
            "f_ref": 20.0,
            "spin_conversion_phase": 0.0,
        },
        "wfg_domain": domain_small,
        "data_domain": domain_small,
        "event_data": sample_small,
        "t_ref": 1248242632.0,
    }

    likelihood_kwargs_masked = deepcopy(likelihood_kwargs)
    likelihood_kwargs_masked["frequency_update"] = {
        "maximum_frequency": f_max_new,
    }

    theta = {
        "chirp_mass": 50.0,
        "mass_ratio": 0.8,
        "a_1": 0.3,
        "a_2": 0.4,
        "tilt_1": 1.3,
        "tilt_2": 1.4,
        "phi_12": 0.3,
        "phi_jl": 3.3,
        "theta_jn": 1.2,
        "luminosity_distance": 4000.0,
        "geocent_time": -0.03,
        "dec": -0.3,
        "ra": 1.5,
        "psi": 1.2,
        "phase": 1.4,
    }

    return likelihood_kwargs, likelihood_kwargs_small, likelihood_kwargs_masked, theta


@pytest.mark.parametrize(
    "setup",
    [
        "ufd_setup",
        "mfd_setup",
    ],
)
def test_likelihood_frequency_masking(request, setup):
    likelihood_kwargs, likelihood_kwargs_small, likelihood_kwargs_masked, theta = (
        request.getfixturevalue(setup)
    )

    likelihood = StationaryGaussianGWLikelihood(
        **likelihood_kwargs
    )  # domain: [20, 1024]
    likelihood_small_range = StationaryGaussianGWLikelihood(
        **likelihood_kwargs_small
    )  # domain: [20, 512]
    likelihood_masked = StationaryGaussianGWLikelihood(
        **likelihood_kwargs_masked
    )  # domain: [20, 512]

    l_vals = likelihood.log_likelihood(theta)
    l_vals_small_range = likelihood_small_range.log_likelihood(theta)
    l_vals_masked = likelihood_masked.log_likelihood(theta)

    assert not np.isclose(l_vals, l_vals_small_range)
    assert not np.isclose(l_vals, l_vals_masked)
    assert np.isclose(l_vals_small_range, l_vals_masked)



import numpy as np
import scipy


def get_index_for_elem(arr, elem):
    return (np.abs(arr - elem)).argmin()


def lorentzian_eval(x, f0, A, Q, delta_f=None):
    """
    Evaluates a Lorentzian function at the given frequencies.
    Parameters
    ----------
    x: array_like
        Frequencies at which the Lorentzian is evaluated.
    f0: float
        Center frequency of the Lorentzian.
    A: float
        Amplitude of the Lorentzian.
    Q: float
        Parameter determining the width of the Lorentzian
    delta_f: float, optional
        If given, the Lorentzian is truncated
    Returns
    -------
    array_like
    """
    if f0 == 0 or A < 0:
        return np.zeros_like(x)

    # used to truncate tails of Lorentzian if necessary. Will have no effect, if delta_f sufficiently large
    truncate = (
        np.where(np.abs(x - f0) <= delta_f, 1, np.exp(-np.abs(x - f0) / delta_f))
        if delta_f
        else np.ones_like(x)
    )

    return truncate * A * (f0**4) / ((x * f0) ** 2 + Q**2 * (f0**2 - x**2) ** 2)


def reconstruct_psds_from_parameters(
    parameters_dict, domain, parameterization_settings
):
    """
    Reconstructs the PSDs from the parameters.
    Parameters
    ----------
    parameters_dict : dict
        Dictionary containing the parameters of the PSDs.
    domain : dingo.gw.noise.domain.Domain
        Domain object containing the frequencies at which the PSDs are evaluated.
    parameterization_settings : dict
        Dictionary containing the settings for the parameterization.
    Returns
    -------
    array_like

    """
    smoothen = parameterization_settings.get("smoothen", False)

    xs = parameters_dict["x_positions"]
    ys_list = parameters_dict["y_values"]
    spectral_features_list = parameters_dict["spectral_features"]

    if spectral_features_list.ndim == 2 and ys_list.ndim == 1:
        spectral_features_list = spectral_features_list[np.newaxis]
        ys_list = ys_list[np.newaxis]

    num_psds = ys_list.shape[0]
    assert num_psds == spectral_features_list.shape[0]

    sigma = parameterization_settings["sigma"]
    frequencies = domain.sample_frequencies

    num_spectral_segments = spectral_features_list.shape[1]
    frequency_segments = np.array_split(
        np.arange(frequencies.shape[0]), num_spectral_segments
    )

    psds = []
    for i in range(num_psds):
        ys = ys_list[i, :]
        spectral_features = spectral_features_list[i, :, :]
        spline = scipy.interpolate.CubicSpline(xs, ys)
        base_noise = spline(frequencies)

        lorentzians = np.array([])
        for j, seg in enumerate(frequency_segments):
            f0, A, Q = spectral_features[j, :]
            lorentzian = lorentzian_eval(frequencies[seg], f0, A, Q)
            # small amplitudes are not modeled to maintain smoothness
            if np.max(lorentzian) <= 3 * sigma:
                lorentzian = np.zeros_like(frequencies[seg])
            lorentzians = np.concatenate((lorentzians, lorentzian), axis=0)
        assert lorentzians.shape == frequencies.shape

        if smoothen:
            psd = np.exp(base_noise + lorentzians)
        else:
            psd = np.exp(np.random.normal(base_noise + lorentzians, sigma))
        psds.append(psd)
    return np.array(psds)



#!/usr/bin/env python
""" Script to importance sample based on Dingo samples. Based on bilby_pipe data
analysis script. """
import os
import sys

import yaml
from bilby_pipe.input import Input
from bilby_pipe.utils import (
    parse_args,
    logger,
    convert_string_to_dict,
    convert_prior_string_input,
    BilbyPipeError,
)

from dingo.gw.data.event_dataset import EventDataset
from dingo.gw.domains import MultibandedFrequencyDomain
from dingo.pipe.default_settings import IMPORTANCE_SAMPLING_SETTINGS
from dingo.pipe.parser import create_parser
from dingo.gw.result import Result

logger.name = "dingo_pipe"


class ImportanceSamplingInput(Input):
    def __init__(self, args, unknown_args):
        super().__init__(args, unknown_args)

        # Generic initialisation
        self.meta_data = dict()
        self.result = None

        # Admin arguments
        self.ini = args.ini
        self.scheduler = args.scheduler
        # self.periodic_restart_time = args.periodic_restart_time
        self.request_cpus = args.request_cpus_importance_sampling

        # Naming arguments
        self.outdir = args.outdir
        self.label = args.label
        self.result_format = args.result_format

        # Samples to run on
        self.proposal_samples_file = args.proposal_samples_file
        self.event_data_file = args.event_data_file

        # Prior
        self.prior_dict = args.prior_dict
        self.default_prior = "PriorDict"
        self.time_reference = "geocent"
        self.prior_dict_updates = args.prior_dict_updates

        # Choices for running
        self.detectors = args.detectors

        # self.sampler = args.sampler
        # self.sampler_kwargs = args.sampler_kwargs
        # self.sampling_seed = args.sampling_seed

        # Frequencies
        # self.sampling_frequency = args.sampling_frequency
        # self.minimum_frequency = args.minimum_frequency
        # self.maximum_frequency = args.maximum_frequency
        # self.reference_frequency = args.reference_frequency

        # # Waveform, source model and likelihood
        # self.waveform_generator_class = args.waveform_generator
        # self.waveform_approximant = args.waveform_approximant
        # self.catch_waveform_errors = args.catch_waveform_errors
        # self.pn_spin_order = args.pn_spin_order
        # self.pn_tidal_order = args.pn_tidal_order
        # self.pn_phase_order = args.pn_phase_order
        # self.pn_amplitude_order = args.pn_amplitude_order
        # self.mode_array = args.mode_array
        # self.waveform_arguments_dict = args.waveform_arguments_dict
        # self.numerical_relativity_file = args.numerical_relativity_file
        # self.frequency_domain_source_model = args.frequency_domain_source_model
        # self.conversion_function = args.conversion_function
        # self.generation_function = args.generation_function
        # self.likelihood_type = args.likelihood_type
        # self.reference_frame = args.reference_frame
        # self.time_reference = args.time_reference
        # self.extra_likelihood_kwargs = args.extra_likelihood_kwargs
        # self.enforce_signal_duration = args.enforce_signal_duration
        #
        # # ROQ
        # self.roq_folder = args.roq_folder
        # self.roq_scale_factor = args.roq_scale_factor
        #
        # Calibration
        self.calibration_model = args.calibration_model
        self.spline_calibration_nodes = args.spline_calibration_nodes
        self.spline_calibration_envelope_dict = args.spline_calibration_envelope_dict
        self.spline_calibration_curves = args.spline_calibration_curves
        self.calibration_correction_type = args.calibration_correction_type

        # # Marginalization
        # self.distance_marginalization = args.distance_marginalization
        # self.distance_marginalization_lookup_table = None
        # self.phase_marginalization = args.phase_marginalization
        # self.time_marginalization = args.time_marginalization
        # self.jitter_time = args.jitter_time

        self._load_proposal()
        self._load_event()  # Must be called after _load_proposal().
        self.importance_sampling_settings = args.importance_sampling_settings

    @property
    def request_memory(self):
        return self.inputs.request_memory_importance_sampling

    def _load_proposal(self):
        self.result = Result(file_name=self.proposal_samples_file)
        if "log_prob" not in self.result.samples.columns:
            raise KeyError(
                "log_prob is not present in proposal samples. This is "
                "required for importance sampling."
            )

    def _load_event(self):
        event_dataset = EventDataset(file_name=self.event_data_file)
        self.result.reset_event(event_dataset)

    @property
    def calibration_marginalization_kwargs(self):
        if self.calibration_model == "CubicSpline":
            return {
                "calibration_envelope": self.spline_calibration_envelope_dict,
                "num_calibration_nodes": self.spline_calibration_nodes,
                "num_calibration_curves": self.spline_calibration_curves,
                "correction_type": self.calibration_correction_type,
            }
        elif self.calibration_model == None:
            return None
        else:
            raise ValueError(
                "The only calibration model which is supported is 'CubicSpline'"
            )

    @property
    def importance_sampling_settings(self):
        return self._importance_sampling_settings

    @importance_sampling_settings.setter
    def importance_sampling_settings(self, settings):
        # Set up defaults.
        if "phase" not in self.result.samples.columns:
            self._importance_sampling_settings = IMPORTANCE_SAMPLING_SETTINGS[
                "PhaseRecoveryDefault"
            ]
        else:
            self._importance_sampling_settings = dict()

        if isinstance(self.result.domain, MultibandedFrequencyDomain):
            self._importance_sampling_settings.update(
                IMPORTANCE_SAMPLING_SETTINGS["MultibandingDefault"]
            )

        if settings is not None:
            if settings.lower() == "default":
                pass
            elif settings.lower() == "phaserecoverydefault":
                self._importance_sampling_settings.update(
                    IMPORTANCE_SAMPLING_SETTINGS["PhaseRecoveryDefault"]
                )
            else:
                self._importance_sampling_settings.update(
                    convert_string_to_dict(settings)
                )
            if "phase_marginalization" in self._importance_sampling_settings:
                self._importance_sampling_settings.pop("synthetic_phase", None)
        else:
            self._importance_sampling_settings = dict()

    def run_sampler(self):
        self.result.use_base_domain = self.importance_sampling_settings.get(
            "use_base_domain", False
        )

        if self.prior_dict_updates:
            logger.info("Updating prior from network prior. Changes:")
            logger.info(
                yaml.dump(
                    self.prior_dict_updates,
                    default_flow_style=False,
                    sort_keys=False,
                )
            )
            self.result.update_prior(self.prior_dict_updates)

        if "synthetic_phase" in self.importance_sampling_settings:
            logger.info("Sampling synthetic phase.")
            synthetic_phase_kwargs = {
                **self.importance_sampling_settings["synthetic_phase"],
                "num_processes": self.request_cpus,
            }
            self.result.sample_synthetic_phase(synthetic_phase_kwargs)

        self.result.importance_sample(
            num_processes=self.request_cpus,
            time_marginalization_kwargs=self.importance_sampling_settings.get(
                "time_marginalization"
            ),
            phase_marginalization_kwargs=self.importance_sampling_settings.get(
                "phase_marginalization"
            ),
            calibration_marginalization_kwargs=self.calibration_marginalization_kwargs,
        )

        self.result.print_summary()
        self.result.to_file(os.path.join(self.result_directory, self.label + ".hdf5"))

    @property
    def priors(self):
        """Read in and compose the prior at run-time"""
        if getattr(self, "_priors", None) is None:
            self._priors = self._get_priors(add_time=False)
        return self._priors

    @property
    def prior_dict_updates(self):
        """The input prior_dict from the ini (if given)

        Note, this is not the bilby prior (see self.priors for that), this is
        a key-val dictionary where the val's are strings which are converting
        into bilby priors in `_get_prior
        """
        return self._prior_dict_updates

    @prior_dict_updates.setter
    def prior_dict_updates(self, prior_dict_updates):
        if isinstance(prior_dict_updates, dict):
            prior_dict_updates = prior_dict_updates
        elif isinstance(prior_dict_updates, str):
            prior_dict_updates = convert_prior_string_input(prior_dict_updates)
        elif prior_dict_updates is None:
            self._prior_dict_updates = None
            return
        else:
            raise BilbyPipeError(
                f"prior_dict_updates={prior_dict_updates} not " f"understood"
            )

        self._prior_dict_updates = {
            self._convert_prior_dict_key(key): val
            for key, val in prior_dict_updates.items()
        }


def create_sampling_parser():
    """Data analysis parser creation"""
    return create_parser(top_level=False)


def main():
    """Data analysis main logic"""
    args, unknown_args = parse_args(sys.argv[1:], create_sampling_parser())
    # log_version_information()
    analysis = ImportanceSamplingInput(args, unknown_args)
    analysis.run_sampler()
    sys.exit(0)



from dingo.gw.domains import (
    Domain,
    UniformFrequencyDomain,
    MultibandedFrequencyDomain,
    TimeDomain,
)


def build_domain(settings: dict) -> Domain:
    """
    Instantiate a domain class from settings.

    Parameters
    ----------
    settings : dict
        Dictionary with 'type' key denoting the type of domain, and keys corresponding
        to the kwargs needed to construct the Domain.

    Returns
    -------
    A Domain instance of the correct type.
    """
    if "type" not in settings:
        raise ValueError(
            f'Domain settings must include a "type" key. Settings included '
            f"the keys {settings.keys()}."
        )

    # The settings other than 'type' correspond to the kwargs of the Domain constructor.
    kwargs = {k: v for k, v in settings.items() if k != "type"}
    if settings["type"] in ["UniformFrequencyDomain", "FrequencyDomain", "FD"]:
        return UniformFrequencyDomain(**kwargs)
    elif settings["type"] in ["MultibandedFrequencyDomain", "MFD"]:
        return MultibandedFrequencyDomain(**kwargs)
    elif settings["type"] in ["TimeDomain", "TD"]:
        return TimeDomain(**kwargs)
    else:
        raise NotImplementedError(f'Domain {settings["type"]} not implemented.')


def build_domain_from_model_metadata(
    model_metadata: dict, base: bool = False
) -> Domain:
    """
    Instantiate a domain class from settings of model.

    Parameters
    ----------
    model_metadata: dict
        model metadata containing information to build the domain
        typically obtained from the model.metadata attribute
    base: bool = False
        If base=True, return domain.base_domain if this is an attribute of domain,
        else return domain. Example: MultibandedFrequencyDomain has a UniformFrequencyDomain
        object as a base_domain. In dingo_pipe, we want to load data in the
        base_domain, and later decimate from base_domain to domain.

    Returns
    -------
    A Domain instance of the correct type.
    """
    domain = build_domain(model_metadata["dataset_settings"]["domain"])
    if "domain_update" in model_metadata["train_settings"]["data"]:
        domain.update(model_metadata["train_settings"]["data"]["domain_update"])
    if base and hasattr(domain, "base_domain"):
        domain = domain.base_domain
    return domain



from pathlib import Path
from typing import get_args

import pytest
import torch

from dingo.core.utils.backward_compatibility import (
    Device,
    torch_available_devices,
    torch_load_with_fallback,
)


class DummyModel(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.linear = torch.nn.Linear(5, 3)


@pytest.fixture
def model_path(tmp_path) -> str:
    model = DummyModel()
    model.linear.weight.data.fill_(1.0)
    model.linear.bias.data.fill_(2.0)
    model_path = tmp_path / "test_model.pt"
    torch.save(model.state_dict(), model_path)
    return str(model_path)


def test_torch_available_device() -> None:
    """
    checks torch_available_devices do not
    raise an error and lists at least cpu
    """
    devices = torch_available_devices()
    assert len(devices) != 0
    assert "cpu" in devices


def test_load_torch_with_fallback(model_path) -> None:
    """
    checks torch_load_with_fallback do not
    raise an error and map at least correctly to cpu
    """
    devices = get_args(Device)
    model, device = torch_load_with_fallback(
        model_path, preferred_map_location="cpu"
    )
    assert device == torch.device("cpu")
    for device in devices:
        # just checking no error is raised.
        # Not checking if it is mapped to the proper device, as we do not
        # want to assume which device is available on the test machine.
        torch_load_with_fallback(model_path, preferred_map_location=device)



from typing import Optional

import numpy as np
import torch

from dingo.gw.gwutils import *
from .base_frequency_domain import BaseFrequencyDomain


class UniformFrequencyDomain(BaseFrequencyDomain):
    """Defines the physical domain on which the data of interest live.

    The frequency bins are assumed to be uniform between [0, f_max] with spacing
    delta_f.

    Given a finite length of time domain data, the Fourier domain data starts
    at a frequency f_min and is zero below this frequency.

    window_kwargs specify windowing used for FFT to obtain FD data from TD data in
    practice.
    """

    def __init__(
        self,
        f_min: float,
        f_max: float,
        delta_f: float,
    ):
        """
        Parameters
        ----------
        f_min : float
        f_max : float
        delta_f : float
        """
        super().__init__()
        self._f_min = f_min
        self._f_max = f_max
        self._delta_f = delta_f
        self._frequency_mask = None

    def update(self, new_settings: dict):
        """
        Update the domain with new settings. This is only allowed if the new settings
        are "compatible" with the old ones. E.g., f_min should be larger than the
        existing f_min.

        Parameters
        ----------
        new_settings : dict
            Settings dictionary. Must contain a subset of the keys contained in
            domain_dict.
        """
        new_settings = new_settings.copy()
        if "type" in new_settings and new_settings.pop("type") not in [
            "UniformFrequencyDomain",
            "FrequencyDomain",
            "FD",
        ]:
            raise ValueError(
                "Cannot update domain to type other than UniformFrequencyDomain."
            )
        for k, v in new_settings.items():
            if k not in ["f_min", "f_max", "delta_f"]:
                raise KeyError(f"Invalid key for domain update: {k}.")
            if k == "delta_f" and v != self._delta_f:
                raise ValueError("Cannot update delta_f.")
        self._set_new_range(
            f_min=new_settings.get("f_min", None), f_max=new_settings.get("f_max", None)
        )

    def _set_new_range(
        self, f_min: Optional[float] = None, f_max: Optional[float] = None
    ) -> None:
        """
        Set a new [f_min, f_max] range for the domain. Both endpoints must be in the
        existing sample_frequencies, and f_min < f_max. Neither endpoint may lie
        outside the current range.
        """
        new_min = f_min if f_min is not None else self.f_min
        new_max = f_max if f_max is not None else self.f_max

        if new_min >= new_max:
            raise ValueError("f_min must be strictly less than f_max.")

        if not (self.f_min <= new_min and new_max <= self.f_max):
            raise ValueError(
                f"Requested range [{new_min}, {new_max}] lies outside "
                f"original range [{self.f_min}, {self.f_max}]."
            )

        missing = [x for x in (new_min, new_max) if x not in self.sample_frequencies]
        if missing:
            raise ValueError(f"Endpoints {missing} not in existing sample_frequencies.")

        self.f_min, self.f_max = new_min, new_max

    def update_data(self, data: np.ndarray, axis: int = -1, low_value: float = 0.0):
        """
        Adjusts data to be compatible with the domain:

        * Below f_min, it sets the data to low_value (typically 0.0 for a waveform,
          but for a PSD this might be a large value).
        * Above f_max, it truncates the data array.

        Parameters
        ----------
        data : np.ndarray
            Data array
        axis : int
            Which data axis to apply the adjustment along.
        low_value : float
            Below f_min, set the data to this value.

        Returns
        -------
        np.ndarray
            The new data array.
        """
        sl = [slice(None)] * data.ndim

        # First truncate beyond f_max.
        sl[axis] = slice(0, self.max_idx + 1)
        data = data[tuple(sl)]

        # Set data value below f_min to low_value.
        sl[axis] = slice(0, self.min_idx)
        data[tuple(sl)] = low_value

        return data

    def get_sample_frequencies_astype(
        self, data: np.ndarray | torch.Tensor
    ) -> np.ndarray | torch.Tensor:
        """
        Returns a 1D frequency array compatible with the last index of data array.

        Decides whether array is numpy or torch tensor (and cuda vs cpu), and whether it
        contains the leading zeros below f_min.

        Parameters
        ----------
        data : Union[np.array, torch.Tensor]
            Sample data

        Returns
        -------
        frequency array compatible with last index
        """
        f = super().get_sample_frequencies_astype(data)

        # Whether to include zeros below f_min
        if data.shape[-1] == len(self) - self.min_idx:
            f = f[self.min_idx :]
        elif data.shape[-1] != len(self):
            raise TypeError(
                f"Data with {data.shape[-1]} frequency bins is "
                f"incompatible with domain."
            )

        return f

    def __len__(self):
        """Number of frequency bins in the domain [0, f_max]"""
        return int(self.f_max / self.delta_f) + 1

    @property
    def sample_frequencies(self) -> np.ndarray:
        if self._sample_frequencies is None:
            num_bins = len(self)
            self._sample_frequencies = np.linspace(
                0.0, self.f_max, num=num_bins, endpoint=True, dtype=np.float32
            )
        return self._sample_frequencies

    @property
    def frequency_mask(self) -> np.ndarray:
        """Mask which selects frequency bins greater than or equal to the
        starting frequency"""
        if self._frequency_mask is None:
            self._frequency_mask = self.sample_frequencies >= self.f_min
        return self._frequency_mask

    def _reset_caches(self):
        self._sample_frequencies = None
        self._sample_frequencies_torch = None
        self._sample_frequencies_torch_cuda = None
        self._frequency_mask = None

    @property
    def frequency_mask_length(self) -> int:
        """Number of samples in the subdomain domain[frequency_mask]."""
        mask = self.frequency_mask
        return len(np.flatnonzero(np.asarray(mask)))

    @property
    def min_idx(self) -> int:
        return round(self._f_min / self._delta_f)

    @property
    def max_idx(self) -> int:
        return round(self._f_max / self._delta_f)

    @property
    def f_max(self) -> float:
        """The maximum frequency [Hz] is typically set to half the sampling
        rate."""
        return self._f_max

    @f_max.setter
    def f_max(self, value):
        self._f_max = float(value)
        self._reset_caches()

    @property
    def f_min(self) -> float:
        """The minimum frequency [Hz]."""
        return self._f_min

    @f_min.setter
    def f_min(self, value):
        self._f_min = float(value)
        self._reset_caches()

    @property
    def delta_f(self) -> float:
        """The frequency spacing of the uniform grid [Hz]."""
        return self._delta_f

    @delta_f.setter
    def delta_f(self, value):
        self._delta_f = float(value)
        self._reset_caches()

    @property
    def duration(self) -> float:
        """Waveform duration in seconds."""
        return 1.0 / self.delta_f

    @property
    def sampling_rate(self) -> float:
        return 2.0 * self.f_max

    @property
    def domain_dict(self):
        """Enables to rebuild the domain via calling build_domain(domain_dict)."""
        return {
            "type": "UniformFrequencyDomain",
            "f_min": self.f_min,
            "f_max": self.f_max,
            "delta_f": self.delta_f,
        }



import copy
import math
import time
from pathlib import Path
from typing import Optional, Union
import sys

import numpy as np
import pandas as pd
import torch
from torchvision.transforms import Compose

from dingo.core.posterior_models import BasePosteriorModel
from dingo.core.result import Result
from dingo.core.result import DATA_KEYS as RESULT_DATA_KEYS
from dingo.core.utils import torch_detach_to_cpu, IterationTracker

# FIXME: transform below should be in core
from dingo.gw.transforms import SelectStandardizeRepackageParameters

#
# Sampler classes are motivated by the approach of Bilby.
#


class Sampler(object):
    """
    Sampler class that wraps a PosteriorModel. Allows for conditional and unconditional
    models.

    Draws samples from the model based on (optional) context data.

    This is intended for use either as a standalone sampler, or as a sampler producing
    initial sample points for a GNPE sampler.

    Methods
    -------
    run_sampler
    log_prob
    to_result
    to_hdf5

    Attributes
    ----------
    model : BasePosteriorModel
    inference_parameters : list
    samples : DataFrame
        Samples produced from the model by run_sampler().
    context : dict
    metadata : dict
    event_metadata : dict
    unconditional_model : bool
        Whether the model is unconditional, in which case it is not provided context
        information.
    transform_pre, transform_post : Transform
        Transforms to be applied to data and parameters during inference. These are
        typically implemented in a subclass.
    """

    def __init__(
        self,
        model: BasePosteriorModel,
    ):
        """
        Parameters
        ----------
        model : BasePosteriorModel
        """
        self.model = model
        self.event_metadata = None

        self.metadata = self.model.metadata.copy()
        if self.metadata["train_settings"]["data"].get("unconditional", False):
            self.unconditional_model = True
            # For unconditional models, the context will be stored with the model. It
            # is needed for calculating the likelihood for importance sampling.
            # However, it will not be used when sampling from the model, since it is
            # unconditional.
            self._context = self.model.context
            self.base_model_metadata = self.metadata["base"]
        else:
            self.unconditional_model = False
            self._context = None
            self.base_model_metadata = self.metadata

        self.inference_parameters = self.metadata["train_settings"]["data"][
            "inference_parameters"
        ]

        self.samples = None
        self._build_domain()

        if self.unconditional_model:
            # Must be after _build_domain(), since setter might access domain.
            self.event_metadata = self.model.event_metadata

        # Must be after _build_domain() since transforms can depend on domain.
        self._initialize_transforms()

        self._pesummary_package = "core"
        self._result_class = Result

    @property
    def context(self):
        """Data on which to condition the sampler. For injections, there should be a
        'parameters' key with truth values."""
        return self._context

    @context.setter
    def context(self, value):
        if value is not None and "parameters" in value:
            if self.event_metadata is None:
                self.event_metadata = {}
            self.event_metadata["injection_parameters"] = value.pop("parameters")
        self._context = value

    @property
    def event_metadata(self):
        """Metadata for data analyzed. Can in principle influence any post-sampling
        parameter transformations (e.g., sky position correction), as well as the
        likelihood detector positions."""
        return self._event_metadata

    @event_metadata.setter
    def event_metadata(self, value):
        self._event_metadata = value

    def _initialize_transforms(self):
        self.transform_pre = Compose([])

        # De-standardize data and extract inference parameters. This needs to be here
        # (and not just in subclasses) in order for run_sampler() to properly execute.
        self.transform_post = SelectStandardizeRepackageParameters(
            {"inference_parameters": self.inference_parameters},
            self.metadata["train_settings"]["data"]["standardization"],
            inverse=True,
            as_type="dict",
        )

    def _run_sampler(
        self,
        num_samples: int,
        context: Optional[dict] = None,
    ) -> dict:
        if not self.unconditional_model:
            if context is None:
                raise ValueError("Context required to run sampler.")
            x = context.copy()
            x["parameters"] = {}
            x["extrinsic_parameters"] = {}

            # transforms_pre are expected to transform the data in the same way for each
            # requested sample. We therefore apply pre-processing only once.
            x = self.transform_pre(context)
            # Require a batch dimension for the embedding network.
            x = x.unsqueeze(0)
            x = [x]
        else:
            if context is not None:
                print("Unconditional model. Ignoring context.")
            x = []

        # For a normalizing flow, we get the log_prob for "free" when sampling,
        # so we always include this. For other architectures, it may make sense to
        # have a flag for whether to calculate the log_prob.
        self.model.network.eval()
        with torch.no_grad():
            y, log_prob = self.model.sample_and_log_prob(*x, num_samples=num_samples)

        if not self.unconditional_model:
            # Squeeze the batch dimension added earlier.
            y = y.squeeze(0)
            log_prob = log_prob.squeeze(0)

        samples = self.transform_post({"parameters": y, "log_prob": log_prob})
        result = samples["parameters"]
        result["log_prob"] = samples["log_prob"]
        return result

    def run_sampler(
        self,
        num_samples: int,
        batch_size: Optional[int] = None,
    ):
        """
        Generates samples and stores them in self.samples. Conditions the model on
        self.context if appropriate (i.e., if the model is not unconditional).

        If possible, it also calculates the log_prob and saves it as a column in
        self.samples. When using GNPE it is not possible to obtain the log_prob due to
        the many Gibbs iterations. However, in the case of just one iteration, and when
        starting from a sampler for the proxy, the GNPESampler does calculate the
        log_prob.

        Allows for batched sampling, e.g., if limited by GPU memory. Actual sampling
        for each batch is performed by _run_sampler(), which will differ for Sampler
        and GNPESampler.

        Parameters
        ----------
        num_samples : int
            Number of samples requested.
        batch_size : int, optional
            Batch size for sampler.
        """
        self.samples = None

        print(f"Running sampler to generate {num_samples} samples.")
        t0 = time.time()
        if not self.unconditional_model:
            if self.context is None:
                raise ValueError("Context must be set in order to run sampler.")
            context = self.context
        else:
            context = None

        # Carry out batched sampling by calling _run_sample() on each batch and
        # consolidating the results.
        if batch_size is None:
            batch_size = num_samples
        full_batches, remainder = divmod(num_samples, batch_size)
        samples = [self._run_sampler(batch_size, context) for _ in range(full_batches)]
        if remainder > 0:
            samples.append(self._run_sampler(remainder, context))
        samples = {p: torch.cat([s[p] for s in samples]) for p in samples[0].keys()}
        samples = {k: v.cpu().numpy() for k, v in samples.items()}

        # Apply any post-sampling transformation to sampled parameters (e.g.,
        # correction for t_ref) and represent as DataFrame.
        self._post_process(samples)
        self.samples = pd.DataFrame(samples)
        print(f"Done. This took {time.time() - t0:.1f} s.")
        sys.stdout.flush()

    def log_prob(self, samples: pd.DataFrame | dict) -> np.ndarray:
        """
        Calculate the model log probability at specific sample points.

        Parameters
        ----------
        samples : pd.DataFrame | dict
            Sample points at which to calculate the log probability.

        Returns
        -------
        np.array of log probabilities.
        """
        if self.context is None and not self.unconditional_model:
            raise ValueError("Context must be set in order to calculate log_prob.")

        if isinstance(samples, dict):
            if np.isscalar(list(samples.values())[0]):
                samples = pd.DataFrame([samples])
            else:
                samples = pd.DataFrame(samples)

        # This undoes any post-correction that would have been done to the samples,
        # before evaluating the log_prob. E.g., the t_ref / sky position correction.
        samples = samples.copy()
        self._post_process(samples, inverse=True)

        # Standardize the sample parameters and place on device.
        y = samples[self.inference_parameters].to_numpy()
        standardization = self.metadata["train_settings"]["data"]["standardization"]
        mean = np.array([standardization["mean"][p] for p in self.inference_parameters])
        std = np.array([standardization["std"][p] for p in self.inference_parameters])
        y = (y - mean) / std
        y = torch.from_numpy(y).to(device=self.model.device, dtype=torch.float32)

        if not self.unconditional_model:
            x = self.context.copy()
            x["parameters"] = {}
            x["extrinsic_parameters"] = {}

            # Context is the same for each sample. Expand across batch dimension after
            # pre-processing.
            x = self.transform_pre(self.context)
            x = x.expand(len(samples), *x.shape)  # TODO: Make this more efficient.
            x = [x]
        else:
            x = []

        self.model.network.eval()
        with torch.no_grad():
            log_prob = self.model.log_prob(y, *x)

        log_prob = log_prob.cpu().numpy()
        log_prob -= np.sum(np.log(std))

        # Pre-processing step may have included a log_prob with the samples.
        if "log_prob" in samples:
            log_prob += samples["log_prob"].to_numpy()

        return log_prob

    def _post_process(self, samples: Union[dict, pd.DataFrame], inverse: bool = False):
        pass

    def _build_domain(self):
        self.domain = None

    def to_result(self) -> Result:
        """
        Export samples, metadata, and context information to a Result instance,
        which can be used for saving or, e.g., importance sampling, training an
        unconditional flow, etc.

        Returns
        -------
        Result
        """
        data_dict = {k: getattr(self, k, None) for k in RESULT_DATA_KEYS}
        # *COPY* the metadata to avoid recursion errors when creating new objects.
        data_dict["settings"] = copy.deepcopy(self.metadata)
        return self._result_class(dictionary=data_dict)

    def to_hdf5(self, label="result", outdir="."):
        dataset = self.to_result()
        file_name = label + ".hdf5"
        Path(outdir).mkdir(parents=True, exist_ok=True)
        dataset.to_file(file_name=Path(outdir, file_name))


class GNPESampler(Sampler):
    """
    Base class for GNPE sampler. It wraps a PosteriorModel *and* a standard Sampler for
    initialization. The former is used to generate initial samples for Gibbs sampling.

    A GNPE network is conditioned on additional "proxy" context theta^, i.e.,

    p(theta | theta^, d)

    The theta^ depend on theta via a fixed kernel p(theta^ | theta). Combining these
    known distributions, this class uses Gibbs sampling to draw samples from the joint
    distribution,

    p(theta, theta^ | d)

    The advantage of this approach is that we are allowed to perform any transformation of
    d that depends on theta^. In particular, we can use this freedom to simplify the
    data, e.g., by aligning data to have merger times = 0 in each detector. The merger
    times are unknown quantities that must be inferred jointly with all other
    parameters, and GNPE provides a means to do this iteratively. See
    https://arxiv.org/abs/2111.13139 for additional details.

    Gibbs sampling breaks access to the probability density, so this must be recovered
    through other means. One way is to train an unconditional flow to represent p(theta^
    | d) for fixed d based on the samples produced through the GNPE Gibbs sampling.
    Starting from these, a single Gibbs iteration gives theta from the GNPE network,
    along with the probability density in the joint space. This is implemented in
    GNPESampler provided the init_sampler provides proxies directly and num_iterations
    = 1.

    Attributes (beyond those of Sampler)
    ----------
    init_sampler : Sampler
        Used for providing initial samples for Gibbs sampling.
    num_iterations : int
        Number of Gibbs iterations to perform.
    iteration_tracker : IterationTracker  **not set up**
    remove_init_outliers : float  **not set up**
    """

    def __init__(
        self,
        model: BasePosteriorModel,
        init_sampler: Sampler,
        num_iterations: int = 1,
    ):
        """
        Parameters
        ----------
        model : BasePosteriorModel
        init_sampler : Sampler
            Used for generating initial samples
        num_iterations : int
            Number of GNPE iterations to be performed by sampler.
        """
        self.gnpe_parameters = []  # Should be set in subclass _initialize_transform()

        super().__init__(model)
        self.init_sampler = init_sampler
        self.num_iterations = num_iterations
        self.iteration_tracker = None
        # remove self.remove_init_outliers of lowest log_prob init samples before gnpe
        self.remove_init_outliers = 0.0

    @property
    def init_sampler(self):
        return self._init_sampler

    @init_sampler.setter
    def init_sampler(self, value):
        self._init_sampler = value
        # Copy this so it persists if we delete the init model.
        self.metadata["init_model"] = self._init_sampler.model.metadata.copy()
        if self._init_sampler.unconditional_model:
            self.context = self._init_sampler.context
            self.event_metadata = self._init_sampler.event_metadata

    @property
    def num_iterations(self):
        """The number of GNPE iterations to perform when sampling."""
        return self._num_iterations

    @num_iterations.setter
    def num_iterations(self, value):
        self._num_iterations = value
        self.metadata["num_iterations"] = self._num_iterations

    @property
    def gnpe_proxy_parameters(self):
        return [p + "_proxy" for p in self.gnpe_parameters]

    def _kernel_log_prob(self, samples):
        raise NotImplementedError("To be implemented in subclass.")

    def _run_sampler(
        self,
        num_samples: int,
        context: Optional[dict] = None,
    ) -> dict:
        if context is None:
            raise ValueError("self.context must be set to run sampler.")

        data_ = self.init_sampler.transform_pre(context)

        # TODO: Reimplement outlier removal in IterationTracker? Save setting somewhere.
        if self.remove_init_outliers == 0.0:
            init_samples = self.init_sampler._run_sampler(num_samples, context)
        else:
            if self.num_iterations == 1:
                print(
                    f"Warning: Removing initial outliers, but only carrying out "
                    f"{self.num_iterations} GNPE iteration. This risks biasing "
                    f"results."
                )
            init_samples = self.init_sampler._run_sampler(
                math.ceil(num_samples / (1 - self.remove_init_outliers)), context
            )
            thr = torch.quantile(init_samples["log_prob"], self.remove_init_outliers)
            inds = torch.where(init_samples["log_prob"] >= thr)[0][:num_samples]
            init_samples = {k: v[inds] for k, v in init_samples.items()}

        # We could be starting with either the GNPE parameters *or* their proxies,
        # depending on the nature of the initialization network.

        start_with_proxies = False
        proxy_log_prob = None
        proxies = {}

        if {p + "_proxy" for p in self.gnpe_parameters}.issubset(init_samples.keys()):
            start_with_proxies = True
            proxy_log_prob = init_samples["log_prob"]
            proxies = {k: init_samples[k] for k in self.gnpe_proxy_parameters}
            # proxies is a dict of torch.Tensors, since it came from _run_sampler(),
            # not run_sampler(). Clone it for a later assertion check.
            init_proxies = {k: v.clone() for k, v in proxies.items()}

        x = {"extrinsic_parameters": init_samples, "parameters": {}}

        #
        # Gibbs sample.
        #

        # Saving as an attribute so we can later access the intermediate Gibbs samples.
        self.iteration_tracker = IterationTracker(store_data=True)

        for i in range(self.num_iterations):
            start_time = time.time()

            if start_with_proxies and i == 0:
                x["extrinsic_parameters"] = proxies.copy()
            else:
                x["extrinsic_parameters"] = {
                    k: x["extrinsic_parameters"][k] for k in self.gnpe_parameters
                }

            # TODO: Depending on whether start_with_proxies is True, this might end up
            #  comparing proxies vs gnpe_parameters for the first iteration.
            self.iteration_tracker.update(
                {k: v.cpu().numpy() for k, v in x["extrinsic_parameters"].items()}
            )

            d = data_.clone()
            x["data"] = d.expand(num_samples, *d.shape)

            x = self.transform_pre(x)

            time_sample_start = time.time()
            self.model.network.eval()
            with torch.no_grad():
                if "context_parameters" in x:
                    y, log_prob = self.model.sample_and_log_prob(
                        x["data"], x["context_parameters"]
                    )
                else:
                    y, log_prob = self.model.sample_and_log_prob(x["data"])

            # Squeeze the extra dimension added by sample_and_log_prob(num_samples=1).
            y = y.squeeze(1)
            log_prob = log_prob.squeeze(1)

            time_sample_end = time.time()

            x["parameters"] = y
            x["log_prob"] = log_prob
            x = self.transform_post(x)

            # Extract the proxy parameters from x["extrinsic_parameters"]. These have
            # not been standardized. They are persistent from before sampling took place,
            # since this is when they were placed here and their values should not have
            # changed.
            proxies = {
                p: x["extrinsic_parameters"][p] for p in self.gnpe_proxy_parameters
            }

            print(
                f"it {i}.\tmin pvalue: {self.iteration_tracker.pvalue_min:.3f}"
                f"\tproxy mean: ",
                *[f"{torch.mean(v).item():.5f}" for v in proxies.values()],
                "\tproxy std:",
                *[f"{torch.std(v).item():.5f}" for v in proxies.values()],
                "\ttimes:",
                time_sample_start - start_time,
                time_sample_end - time_sample_start,
                time.time() - time_sample_end,
            )

        #
        # Prepare final result.
        #

        if start_with_proxies and self.num_iterations == 1:
            # In this case it makes sense to save the log_prob and the proxy parameters.

            samples = x["parameters"]
            samples["log_prob"] = x["log_prob"] + proxy_log_prob

            # The log_prob returned by gnpe is not just the log_prob over parameters
            # theta, but instead the log_prob in the *joint* space q(theta,theta^|x),
            # including the proxies theta^. For importance sampling this means,
            # that the target density is
            #
            #       p(theta,theta^|x) = p(theta^|theta) * p(theta|x).
            #
            # We compute log[p(theta^|theta)] below and store it as
            # samples["delta_log_prob_target"], such that for importance sampling we
            # only need to evaluate log[p(theta|x)] and add this correction.

            # Proxies should be sitting in extrinsic_parameters.
            all_params = {**x["extrinsic_parameters"], **samples}
            all_params = {k: torch_detach_to_cpu(v) for k, v in all_params.items()}
            kernel_log_prob = self._kernel_log_prob(all_params)
            samples["delta_log_prob_target"] = torch.Tensor(kernel_log_prob)

        else:
            # Otherwise we only save the inference parameters, and no log_prob.
            # Alternatively we could save the entire chain and the log_prob, but this
            # is not useful for our purposes.

            samples = x["parameters"]

        # Include the proxies along with the inference parameters. The variable proxies
        # gets set at the end of each Gibbs loop.
        samples.update(proxies)

        # Safety check for unconditional flows. Make sure the proxies haven't changed.
        if start_with_proxies and self.num_iterations == 1:
            for k in proxies:
                assert torch.equal(proxies[k], init_proxies[k])

        return samples



import numpy as np
import matplotlib.pyplot as plt
from scipy import linalg
import torch
import torch.nn as nn
import torch.optim as optim
from itertools import chain


########################
# test utils for enet  #
########################


def generate_1d_series_data(batch_size, num_bins, alpha=20, beta=0):
    """Generate a set of batch_size 1d series with num_bins bins."""

    theta = np.random.rand(batch_size, 2)
    x = np.linspace(0, 1, num_bins)
    y = np.zeros((batch_size, num_bins))
    for i in range(batch_size):
        y[i, :] = np.cos(
            x * (0.1 + theta[i, 0]) * alpha + theta[i, 1]) * x ** beta
    y = y * np.exp(2j * np.pi * x ** 2)
    return y


def generate_1d_datasets_and_reduced_basis(batch_size=1000, num_bins=200,
                                           plot=False, n_rb=10):
    """
    Generate two sets of batch_size 1d series with num_bins bins,
    and generate a reduced basis for these.
    """

    # generate datasets
    y1 = generate_1d_series_data(batch_size, num_bins, alpha=20, beta=0)
    y2 = generate_1d_series_data(batch_size, num_bins, alpha=20, beta=1)
    # generate reduced basis
    _, _, Vh1 = linalg.svd(y1)
    V1 = Vh1.T.conj()
    _, _, Vh2 = linalg.svd(y2)
    V2 = Vh2.T.conj()
    # plot data
    if plot:
        plt.plot(y1[0].real)
        plt.plot((y1[0] @ V1[:, :n_rb] @ Vh1[:n_rb, :]).real)
        plt.plot((y1[0] @ V2[:, :n_rb] @ Vh2[:n_rb, :]).real)
        plt.show()
        plt.plot((y1[0] @ V1[:, :n_rb] @ Vh1[:n_rb, :]).real - y1[0].real)
        plt.plot((y1[0] @ V2[:, :n_rb] @ Vh2[:n_rb, :]).real - y1[0].real)
        plt.show()
        plt.plot((y2[0] @ V1[:, :n_rb] @ Vh1[:n_rb, :]).real - y2[0].real)
        plt.plot((y2[0] @ V2[:, :n_rb] @ Vh2[:n_rb, :]).real - y2[0].real)
        plt.show()
    return (y1, y2), (V1, V2)


def get_y_batch(y_list, num_channels=3):
    """
    Turn multiple sets of 1d series into a batch for a nn by concatenation.
    Real and imaginary parts are separated into separate channels.
    """

    y_batch = torch.ones(
        (y_list[0].shape[0], len(y_list), num_channels, y_list[0].shape[1]))
    for idx, y in enumerate(y_list):
        y_batch[:, idx, 0, :] = torch.from_numpy(y.real).float()
        y_batch[:, idx, 1, :] = torch.from_numpy(y.imag).float()
    return y_batch


def project_onto_reduced_basis_and_concat(y_list, V_rb_list, n_rb):
    """
    Projects each y in y_list onto the corresponding reduced basis in V_rb_list.
    n_rb basis elements are used. For each y, the real and imaginary parts
    are concatenated. Finally, the reduced basis representations of all y are
    concatenated and returned.
    """

    projections = [[(y[:] @ V_rb[:, :n_rb]).real, (y[:] @ V_rb[:, :n_rb]).imag]
                   for y, V_rb in zip(y_list, V_rb_list)]
    projections = list(chain.from_iterable(projections))
    return np.concatenate(projections, axis=1)


def check_model_forward_pass(model, expected_output_shape, input_shape=None,
                             batch_size=100, x=None):
    """
    This function tests the forward pass of the model. It generates random
    input x with shape (batch_size, *input_shape) and performs a forward
    pass. The output y = model(x) is checked for the correct output shape,
    y.shape == (batch_size, *expected_output_shape), and whether distinct
    inputs generate distinct outputs.

    :param model: model to be checked
    :param input_shape: shape of the input data, omitting batch dimension
    :param expected_output_shape: expected shape of the output, omitting
    batch dimension
    :param batch_size: batch size
    :param x: input to model, if provided input is not generated
    """
    if x is None:
        assert input_shape is not None, \
            'input_shape required when x not provided.'
        x = torch.rand((batch_size, *input_shape))
    if isinstance(x, tuple):
        y = model(*x)
    else:
        y = model(x)
    # y = forward_pass_with_unpacked_tuple(model,x)  # replaces y = model(x)
    # check output shape
    assert y.shape[1:] == (*expected_output_shape,), \
        'Unexpected shape of model output.'
    # check that results are different for different inputs
    permuted_indices = [idx - 1 for idx in range(y.shape[0])]
    loss_fn = nn.L1Loss(reduction='none')
    loss = loss_fn(y, y[permuted_indices])
    assert torch.median(loss) / torch.median(torch.abs(y)) > 0.01, \
        'Model outputs could be insensitive to input. Check manually.'


def check_model_backward_pass(model, input_shape=None, batch_size=100, x=None):
    """
    This function tests the backward pass of the model and the optimizer
    step. It generates random input x with shape (batch_size, *input_shape),
    performs a forward pass y=model(x), computes the L1 loss between the output
    y and a random target vector with the same shape. This function then
    checks, whether the loss decreases after a backward pass and an optimizer
    step.

    :param model: model to be checked
    :param input_shape: shape of the input data, omitting batch dimension
    :param batch_size: batch size
    :param x: input to model, if provided input is not generated
    """
    if x is None:
        assert input_shape is not None, \
            'input_shape required when x not provided.'
        x = torch.rand((batch_size, *input_shape))
    if isinstance(x, tuple):
        y_0 = model(*x)
    else:
        y_0 = model(x)
    #y_0 = forward_pass_with_unpacked_tuple(model, x) # replaces y_0 = model(x)
    optimizer = optim.Adam(model.parameters(), lr=0.001)
    loss_fn = nn.L1Loss()
    target = torch.rand_like(y_0)
    loss_before = loss_fn(y_0, target)
    loss_before.backward()
    optimizer.step()
    if isinstance(x, tuple):
        y_1 = model(*x)
    else:
        y_1 = model(x)
    # y_1 = forward_pass_with_unpacked_tuple(model, x) # replaces y_1 = model(x)
    loss_after = loss_fn(y_1, target)
    assert loss_after < loss_before, \
        'Loss does not decrease with optimizer step.'
    return y_0, y_1



import torch

from dingo.gw.domains import UniformFrequencyDomain, TimeDomain
from dingo.gw.domains import build_domain
import pytest
import numpy as np


@pytest.fixture
def uniform_FD_params():
    f_min = 20.0
    f_max = 4096.0
    delta_f = 1.0 / 4.0
    return {"f_min": f_min, "f_max": f_max, "delta_f": delta_f}


def test_uniform_FD(uniform_FD_params):
    p = uniform_FD_params
    domain = UniformFrequencyDomain(**p)
    n = int(p["f_max"] / p["delta_f"]) + 1
    frequencies_expected = np.linspace(0, p["f_max"], n)
    frequencies = domain()
    assert np.linalg.norm(frequencies - frequencies_expected) < 1e-15


def test_uniform_FD_mask(uniform_FD_params):
    p = uniform_FD_params
    domain = UniformFrequencyDomain(**p)
    mask = domain.frequency_mask
    n_masked = int((p["f_max"] - p["f_min"]) / p["delta_f"]) + 1
    frequencies_expected_masked = np.linspace(p["f_min"], p["f_max"], n_masked)
    frequencies_masked = domain()[mask]
    assert np.linalg.norm(frequencies_masked - frequencies_expected_masked) < 1e-15


def test_FD_domain_dict(uniform_FD_params):
    p = uniform_FD_params
    domain = UniformFrequencyDomain(**p)
    domain2 = build_domain(domain.domain_dict)
    assert domain.__dict__ == domain2.__dict__


def test_FD_update_data(uniform_FD_params):
    p = uniform_FD_params
    domain = UniformFrequencyDomain(**p)
    n = int(p["f_max"] / p["delta_f"]) + 1
    nt = int((p["f_max"] - p["f_min"]) / p["delta_f"]) + 1
    # test that the sample frequencies are correct
    assert n == len(domain) == len(domain())
    assert nt == int(np.sum(domain.frequency_mask))
    assert np.all(domain() == np.linspace(0, p["f_max"], n))
    assert np.all(
        domain()[np.nonzero(domain.frequency_mask)]
        == np.linspace(p["f_min"], p["f_max"], nt)
    )
    # Test that data range adjustment works specifying different axis
    a = np.random.rand(10, n, 20)
    b = domain.update_data(a, axis=1)
    assert np.all(b[:, : domain.min_idx, :] == 0.0)
    assert np.all(b[:, domain.min_idx : n + 1, :] == a[:, domain.min_idx :, :])
    # Test that data range adjustment works as intended with nonzero low_value
    a = np.random.random(n)
    assert np.all(domain.update_data(a) == a * domain.frequency_mask)
    assert np.all(
        domain.update_data(a, low_value=5.0)[: domain.min_idx]
        == 5.0 * np.ones(round(p["f_min"] / p["delta_f"]))
    )


def test_FD_set_new_range(uniform_FD_params):
    p = uniform_FD_params
    domain = UniformFrequencyDomain(**p)
    # test that ValueErrors are raised for infeasible inputs
    with pytest.raises(ValueError):
        domain._set_new_range(p["f_max"] + 10, None)
    with pytest.raises(ValueError):
        domain._set_new_range(p["f_min"] - 10, None)
    with pytest.raises(ValueError):
        domain._set_new_range(None, p["f_max"] + 10)
    with pytest.raises(ValueError):
        domain._set_new_range(None, p["f_min"] - 10)
    with pytest.raises(ValueError):
        domain._set_new_range(p["f_min"] + 10, p["f_min"] + 5)
    with pytest.raises(ValueError):
        domain._set_new_range(p["f_min"] + 0.1, None)
    with pytest.raises(ValueError):
        domain._set_new_range(None, p["f_max"] - 0.1)
    # test that setting new frequency range works as intended
    f_min_new, f_max_new = 40, 800
    n = int(f_max_new / p["delta_f"]) + 1
    domain._set_new_range(f_min_new, f_max_new)
    assert n == len(domain) == len(domain())
    assert np.all(domain() == np.linspace(0, f_max_new, n))
    mask = np.ones(n)
    mask[: round(f_min_new / p["delta_f"])] = 0.0
    assert np.all(domain.frequency_mask == mask)
    # Test that data range adjustment works as intended when setting new fmin, fmax
    a = np.random.random(n)
    assert np.all(domain.update_data(a) == a * domain.frequency_mask)
    assert np.all(
        domain.update_data(a, low_value=5.0)[: domain.min_idx]
        == 5.0 * np.ones(round(f_min_new / p["delta_f"]))
    )


def test_FD_time_translation(uniform_FD_params):
    p = uniform_FD_params
    domain = UniformFrequencyDomain(**p)
    # test normal time translation
    dt = 1e-3
    a = np.sin(np.outer(np.arange(3) + 1, domain()) / 100) + 1j * np.cos(
        np.outer(np.arange(3) + 1, domain()) / 100
    )
    a_dt = a * np.exp(-2j * np.pi * domain() * dt)
    b_dt = domain.time_translate_data(a, dt)
    assert np.allclose(a_dt, b_dt)
    assert not np.allclose(a_dt, domain.time_translate_data(a, 1.01 * dt))
    # test that time translation can be undone
    assert np.allclose(a, domain.time_translate_data(a_dt, -dt))
    assert not np.allclose(a, domain.time_translate_data(a_dt, -1.01 * dt))


def test_FD_time_translation_torch(uniform_FD_params):
    p = uniform_FD_params
    domain = UniformFrequencyDomain(**p)
    batch_size = 5
    num_detectors = 2
    dt = torch.randn(batch_size, num_detectors, dtype=torch.float32)
    data = torch.empty(
        (batch_size, num_detectors, 3, len(domain) - domain.min_idx),
        dtype=torch.float32,
    )
    constant_value = 2.0
    f = domain.sample_frequencies_torch[domain.min_idx :]
    for i in range(batch_size):
        for j in range(num_detectors):
            signal = torch.exp(-2j * np.pi * dt[i, j] * f)
            data[i, j, 0, :] = signal.real
            data[i, j, 1, :] = signal.imag
            data[i, j, 2, :] = constant_value
    result = domain.time_translate_data(data, -dt)
    assert result.dtype == torch.float32
    assert result.shape == data.shape
    assert torch.allclose(result[..., 0, :], torch.tensor(1.0), atol=1e-4)
    # Tolerance of 1e-2 is required in the imaginary part, likely because we are
    # checking that it *vanishes*. This is a consequence of single-precision floats.
    # TODO: Is there a way to improve on this?
    assert torch.allclose(result[..., 1, :], torch.tensor(0.0), atol=1e-2)
    assert torch.allclose(result[..., 2, :], torch.tensor(constant_value))


def test_FD_caching(uniform_FD_params):
    p = uniform_FD_params
    domain = UniformFrequencyDomain(**p)
    domain_ref = UniformFrequencyDomain(**p)

    assert np.all(domain() == domain_ref())
    # we now modify domain._f_max by hand, which should not be done, as this
    # does not update the cached properties
    domain._f_max = 50
    assert np.all(domain() == domain_ref())
    domain._reset_caches()
    # after clearing the cache, the __call__ method should return the correct
    # result
    assert len(domain()) < len(domain_ref())




def _strip_unwanted_submission_keys(job):
    job.getenv = None
    job.universe = None
    extra_lines = []
    for line in job.extra_lines:
        if not line.startswith("priority") and not line.startswith("accounting_group") and not line.startswith("ENV GET HTGETTOKENOPTS"):
            extra_lines.append(line)
    job.extra_lines = extra_lines

def dict_to_string(d):
    """Convert a dictionary to a string representation."""
    return "{"  + ", ".join(f"{key}:{value}" for key, value in d.items()) + "}"


from os.path import join
import numpy as np
import math
import pandas as pd
import matplotlib.pyplot as plt
from dingo.core.utils.plotting import plot_corner_multi
from dingo.gw.result import Result


def plot_posterior_slice2d(
    sampler, theta, theta_range, n_grid=100, num_processes=1, outname=None
):
    # prepare theta grid
    keys = list(theta_range.keys())
    axis0 = np.linspace(theta_range[keys[0]][0], theta_range[keys[0]][1], n_grid)
    axis1 = np.linspace(theta_range[keys[1]][0], theta_range[keys[1]][1], n_grid)
    full_axis0, full_axis1 = np.meshgrid(axis0, axis1)

    p0, p1 = theta[keys[0]], theta[keys[1]]
    theta_grid = pd.DataFrame(
        np.repeat(np.array(theta)[np.newaxis], n_grid**2, axis=0),
        columns=theta.keys(),
    )
    theta_grid[keys[0]] = full_axis0.flatten()
    theta_grid[keys[1]] = full_axis1.flatten()

    # compute log_prob
    log_probs_target = (
        sampler.likelihood.log_likelihood_multi(theta_grid, num_processes)
        # + sampler.prior.ln_prob(theta_grid, axis=0)
        - sampler.log_evidence
    )
    log_probs_target -= np.max(log_probs_target)
    plt.clf()
    plt.imshow(np.exp(log_probs_target).reshape((n_grid, n_grid)), cmap="viridis")
    plt.colorbar()
    ticks_positions = np.arange(0, n_grid, n_grid // 5)
    plt.xticks(ticks_positions, labels=[f"{v:.2f}" for v in axis0[ticks_positions]])
    plt.xlabel(keys[0])
    plt.yticks(ticks_positions, labels=[f"{v:.2f}" for v in axis1[ticks_positions]])
    plt.ylabel(keys[1])

    if outname is not None:
        plt.savefig(outname)
    else:
        plt.show()


def plot_posterior_slice(
    sampler,
    theta,
    theta_range,
    outname=None,
    num_processes=1,
    n_grid=200,
    parameters=None,
    normalize_conditionals=False,
):
    # Put a cap on number of processes to avoid overhead.
    # num_processes = min(num_processes, n_grid // 50)

    # repeat theta n_grid times
    theta_grid = pd.DataFrame(
        np.repeat(np.array(theta)[np.newaxis], n_grid, axis=0),
        columns=theta.keys(),
    )
    plt.clf()
    if parameters is None:
        parameters = theta.keys()
    # rows, columns = 3, 5
    columns = math.ceil(np.sqrt(len(parameters)))
    rows = math.ceil(len(parameters) / columns)
    fig, ax = plt.subplots(
        rows, columns, figsize=(columns * 6, rows * 4), squeeze=False
    )
    for idx, param in enumerate(parameters):
        # axis with scan for param
        param_axis = np.linspace(theta_range[param][0], theta_range[param][1], n_grid)
        # build theta_grid for param
        theta_param = theta_grid.copy()
        theta_param[param] = param_axis
        # evaluate the posterior at theta_grid
        log_probs_target = (
            sampler.likelihood.log_likelihood_multi(theta_param, num_processes)
            + sampler.prior.ln_prob(theta_param, axis=0)
            - sampler.log_evidence
        )
        # evaluate nde at theta_grid
        # log_probs_proposal = get_log_probs_from_proposal(nde, theta_param)
        log_probs_proposal = sampler.log_prob(theta_param)

        # plot
        target = np.exp(log_probs_target)
        proposal = np.exp(log_probs_proposal)
        if normalize_conditionals:
            target /= target.sum()
            proposal /= proposal.sum()
        i, j = idx // columns, idx % columns
        ax[i, j].set_xlabel(param)
        ax[i, j].axvline([theta[param]], color="black", label="theta")
        ax[i, j].plot(param_axis, target, label="target")
        ax[i, j].plot(param_axis, proposal, label="proposal")

    plt.legend()
    if outname is not None:
        plt.savefig(outname)
    else:
        plt.show()


def plot_diagnostics(
    result: Result,
    outdir,
    num_processes=1,
    num_slice_plots=0,
    n_grid_slice1d=200,
    n_grid_slice2d=100,
    params_slice2d=None,
):
    theta = result.samples
    theta = theta.drop(columns="delta_log_prob_target", errors="ignore")
    weights = theta["weights"].to_numpy()
    log_probs_proposal = theta["log_prob"].to_numpy()

    log_prior = theta["log_prior"].to_numpy()
    log_likelihood = theta["log_likelihood"].to_numpy()
    log_probs_target = log_prior + log_likelihood

    n_eff = result.n_eff
    sample_efficiency = result.sample_efficiency
    log_evidence = result.log_evidence
    log_evidence_std = result.log_evidence_std

    # Plot weights
    plt.clf()
    y = weights / np.mean(weights)
    x = log_probs_proposal
    plt.xlabel("proposal log_prob")
    plt.ylabel("Weights (normalized to mean 1)")
    y_lower = 1e-4
    y_upper = math.ceil(
        np.max(y) / 10 ** math.ceil(np.log10(np.max(y)) - 1)
    ) * 10 ** math.ceil(np.log10(np.max(y)) - 1)
    plt.ylim(y_lower, y_upper)
    n_below = len(np.where(y < y_lower)[0])
    plt.yscale("log")
    plt.title(
        f"IS Weights. {n_below} below {y_lower}. "
        f"Effective samples: {n_eff:.0f} (Efficiency = {100 * sample_efficiency:.2f}%)."
    )
    plt.scatter(x, y, s=0.5)
    plt.savefig(join(outdir, "weights.png"))

    # plot log probs
    plt.clf()
    x = log_probs_proposal
    y = log_probs_target - log_evidence
    plt.xlabel("NDE log_prob (proposal)")
    plt.ylabel("Posterior log_prob (target) - log_evidence")
    y_lower, y_upper = np.max(y) - 20, np.max(y)
    plt.ylim(y_lower, y_upper)
    n_below = len(np.where(y < y_lower)[0])
    plt.title(
        f"Target log_probs. {n_below} below {y_lower:.2f}. "
        f"Log_evidence {log_evidence:.3f} +- {log_evidence_std:.3f}."
    )
    plt.scatter(x, y, s=0.5)
    plt.plot([y_upper - 20, y_upper], [y_upper - 20, y_upper], color="black")
    plt.savefig(join(outdir, "log_probs.png"))

    # # plot slice plots
    # if num_slice_plots > 0:
    #     theta_slice_plots = theta.sample(num_slice_plots).drop(
    #         columns=["weights", "log_prob", "log_prior", "log_likelihood"]
    #     )
    #     # global range for 1D parameter scan
    #     theta_range_1d = {
    #         k: (np.min(theta[k]), np.max(theta[k])) for k in theta_slice_plots.columns
    #     }
    #     # generate slice plots for each theta sample
    #     for idx, (_, theta_idx) in enumerate(theta_slice_plots.iterrows()):
    #         # 1d slice plots
    #         plot_posterior_slice(
    #             sampler,
    #             theta_idx,
    #             theta_range_1d,
    #             num_processes=num_processes,
    #             n_grid=n_grid_slice1d,
    #             outname=join(outdir, f"theta_{idx}_posterior_slice1d.pdf"),
    #         )
    #         # optionally, plot 2d slice plots
    #         if params_slice2d is not None:
    #             # Get parameter ranges for 2d scan.
    #             # We set this as a 1 std area around the respective parameter values,
    #             # except for the phase which we scan in [0, 2pi].
    #             stds = {k: np.std(theta[k]) for k in theta.keys()}
    #             theta_range_2d = {
    #                 k: (theta_idx[k] - stds[k], theta_idx[k] + stds[k])
    #                 for k in theta_idx.keys()
    #             }
    #             theta_range_2d["phase"] = (0, 2 * np.pi)
    #             for param_pair in params_slice2d:
    #                 plot_posterior_slice2d(
    #                     sampler,
    #                     theta_idx,
    #                     {k: theta_range_2d[k] for k in param_pair},
    #                     num_processes=num_processes,
    #                     n_grid=n_grid_slice2d,
    #                     outname=join(
    #                         outdir,
    #                         f"theta_{idx}_posterior_slice2d_"
    #                         f"{param_pair[0]}-{param_pair[1]}.pdf",
    #                     ),
    #                 )

    # cornerplot with unweighted vs. weighted samples
    weights = weights / np.mean(weights)
    threshold = 1e-3
    inds = np.where(weights > threshold)[0]
    theta_new = theta.loc[inds]
    weights_new = weights[inds]
    print(
        f"Generating cornerplot with {len(theta_new)} out of {len(theta)} IS samples."
    )

    plot_corner_multi(
        [theta[: len(theta_new)], theta_new],
        weights=[None, weights_new],
        labels=["Dingo", "Dingo-IS"],
        filename=join(outdir, "cornerplot_dingo-is.pdf"),
    )
